import numpy as np
from ase.build import molecule
from ase.constraints import FixedPlane
from ase.optimize import QuasiNewton
from gpaw import GPAW, FermiDirac

# Initialization
molname = 'benzene-mol'
dimername = 'benzene-dimer'
f = open('benzene-dimer-T-shape.dat', 'w')
h = 0.18
xc = 'vdW-DF'

# relaxation of the benzene molecule
benz = molecule('C6H6')
benz.set_pbc(False)
tags = np.zeros_like(benz)
benz.set_tags(tags)
benz.center(vacuum=4.0)
cell = benz.get_cell()

calc = GPAW(nbands=-1,
            h=h,
            xc=xc,
            occupations=FermiDirac(0.0),
            txt=molname + '_relax.txt')
benz.calc = calc

# qn constraint
for i in range(len(benz)):
    plane = FixedPlane(i, (0, 0, 1))
    benz.set_constraint(plane)

qn = QuasiNewton(benz,
                 logfile=molname + '_relax.log',
                 trajectory=molname + '_relax.traj')
qn.run(fmax=0.01)

e_mol = benz.get_potential_energy()
del calc

# mapping out the benzene dimer (T-shaped) intermolecular distance

e_array = np.zeros(20)
d_array = np.zeros(20)

k = 0
for i in np.linspace(-6, 6, 20):
    k_str = str(k)
    z = 6.0 + i * 0.3
    BENZ = benz.copy()
    dimer = BENZ.copy()
    tags = np.ones_like(dimer)
    dimer.set_tags(tags)
    BENZ.rotate(90, 'x', center='COM')
    BENZ.translate([0, 0, z])
    dimer.extend(BENZ)
    dimer.set_cell([cell[0, 0], cell[1, 1], cell[2, 2] + 8])
    dimer.center()
    dimer.set_pbc(False)
    pos = dimer.get_positions()
    d = pos[21, 2] - pos[0, 2]

    calc = GPAW(nbands=-2,
                h=h,
                xc=xc,
                occupations=FermiDirac(0.0),
                txt=dimername + '_' + k_str + '.txt')
    dimer.calc = calc
    e_dimer = dimer.get_potential_energy()
    del calc

    # interaction energy
    e_int = e_dimer - 2 * e_mol
    e_array[k] = e_int
    d_array[k] = d

    print(str(round(d, 3)), e_int, file=f)
    f.flush()
    k += 1

# E_int-curve minimum
e_0 = 100.
d_0 = 0.
for i in range(len(e_array)):
    if e_array[i] < e_0:
        e_0 = e_array[i]
        d_0 = d_array[i]
print('****************', file=f)
print('Minimum (E_int,d):', e_0, d_0, file=f)
f.close()

assert abs(e_0 - -0.11) < 0.01
assert abs(d_0 - 2.86) < 0.05



from myqueue.task import task


def create_tasks():
    return [
        task('top.py@8:15m'),
        task('pdos.py', deps='top.py'),
        task('lcaodos_gs.py@8:15m'),
        task('lcaodos_plt.py', deps='lcaodos_gs.py')]



from ase import Atoms
from ase.units import Hartree
from gpaw import GPAW, PoissonSolver, FermiDirac, Davidson, MixerSum
from gpaw.test import equal


def test_xc_revPBE_Li():
    a = 5.0
    n = 24
    li = Atoms('Li', magmoms=[1.0], cell=(a, a, a), pbc=True)

    calc = GPAW(gpts=(n, n, n), nbands=1, xc=dict(name='oldPBE', stencil=1),
                poissonsolver=PoissonSolver(),
                mixer=MixerSum(0.6, 5, 10.0),
                eigensolver=Davidson(4),
                convergence=dict(eigenstates=4.5e-8),
                occupations=FermiDirac(0.0))
    li.calc = calc
    e = li.get_potential_energy() + calc.get_reference_energy()
    equal(e, -7.462 * Hartree, 1.4)

    calc.set(xc=dict(name='revPBE', stencil=1))
    erev = li.get_potential_energy() + calc.get_reference_energy()

    equal(erev, -7.487 * Hartree, 1.3)
    equal(e - erev, 0.025 * Hartree, 0.002 * Hartree)

    print(e, erev)
    energy_tolerance = 0.002
    equal(e, -204.381098849, energy_tolerance)  # svnversion 5252
    equal(erev, -205.012303379, energy_tolerance)  # svnversion 5252



import os

scalapack = True
fftw = True

# Clean out any autodetected things, we only want the EasyBuild
# definitions to be used.
libraries = ['openblas', 'fftw3', 'readline', 'gfortran']
mpi_libraries = []
include_dirs = []

# Use EasyBuild scalapack from the active toolchain
libraries += ['scalapack']

# Use EasyBuild libxc
libxc = os.getenv('EBROOTLIBXC')
if libxc:
    include_dirs.append(os.path.join(libxc, 'include'))
    libraries.append('xc')

# libvdwxc:
# Use EasyBuild libvdwxc
# This will only work with the foss toolchain.
libvdwxc = os.getenv('EBROOTLIBVDWXC')
if libvdwxc:
    include_dirs.append(os.path.join(libvdwxc, 'include'))
    libraries.append('vdwxc')

# ELPA:
# Use EasyBuild ELPA if loaded
elpa = os.getenv('EBROOTELPA')
if elpa:
    libraries += ['elpa']
    elpaversion = os.path.basename(elpa).split('-')[0]
    library_dirs = [os.path.join(elpa, 'lib')]
    extra_link_args = [f'-Wl,-rpath={elpa}/lib']
    include_dirs.append(os.path.join(elpa, 'include', 'elpa-' + elpaversion))

# Now add a EasyBuild "cover-all-bases" library_dirs
library_dirs = os.getenv('LD_LIBRARY_PATH').split(':')



from myqueue.task import task


def create_tasks():
    return [
        task('timepropagation_calculate.py@8:1h'),
        task('timepropagation_continue.py@8:1h',
             deps='timepropagation_calculate.py'),
        task('timepropagation_postprocess.py@8:5m',
             deps='timepropagation_continue.py'),
        task('timepropagation_plot.py@1:5m',
             deps='timepropagation_postprocess.py'),
        task('casida_calculate.py@8:1h'),
        task('casida_postprocess.py@8:5m', deps='casida_calculate.py'),
        task('casida_plot.py@1:5m', deps='casida_postprocess.py')]



scalapack = False

extra_compile_args = ['-O3', '-std=c99', '-fpic']

compiler = '/afs/crc.nd.edu/x86_64_linux/openmpi/1.3.2/gnu/bin/mpicc'

libraries = ['mkl_def', 'mkl_lapack', 'mkl_core', 'mkl_sequential', 'mkl_gf_lp64', 'iomp5']

mkl_lib_path = '/opt/crc/scilib/mkl/10.1.0.015/lib/em64t/'
ompi_lib_path = '/afs/crc.nd.edu/x86_64_linux/openmpi/1.3.2/gnu/lib/'

library_dirs = [mkl_lib_path, ompi_lib_path]

extra_link_args =['-Wl,-rpath='+mkl_lib_path+',-rpath='+ompi_lib_path]

define_macros += [('GPAW_NO_UNDERSCORE_CBLACS', '1')]
define_macros += [('GPAW_NO_UNDERSCORE_CSCALAPACK', '1')]



from ase.build import bulk
from gpaw import GPAW, FermiDirac
from gpaw.wavefunctions.pw import PW

pwcutoff = 400.0
# Plane wave cutoff

k = 6
# NxNxN k-point sampling, gamma-centred grid

alat = 5.421
# Si lattice constant


# Do the bulk calculation


bulk_crystal = bulk('Si', 'diamond', a=alat)
bulk_calc = GPAW(mode=PW(pwcutoff),
                 kpts={'size': (k, k, k), 'gamma': True},
                 xc='PBE',
                 txt='si.pbe_output.txt',
                 parallel={'band': 1},
                 occupations=FermiDirac(0.01)
                 )

bulk_crystal.calc = bulk_calc
e0_bulk_pbe = bulk_crystal.get_potential_energy()



from myqueue.task import task


def create_tasks():
    return [
        task('xyz.py'),
        task('magnetism1.py', tmax='1h', deps='xyz.py'),
        task('magnetism2.py', tmax='2h', deps='xyz.py')]



from myqueue.task import task


def create_tasks():
    return [
        task('gs_Si.py@4:20m'),
        task('eps_Si.py@4:6h', deps='gs_Si.py'),
        task('plot_Si.py@1:10m', deps='eps_Si.py'),
        task('gs_MoS2.py@4:1h'),
        task('pol_MoS2.py@64:33h', deps='gs_MoS2.py'),
        task('plot_MoS2.py@1:10m', deps='pol_MoS2.py'),
        task('get_2d_eps.py@1:8h', deps='gs_MoS2.py'),
        task('plot_2d_eps.py@1:10m', deps='get_2d_eps.py'),
        task('alpha_MoS2.py@1:10m', deps='gs_MoS2.py')]



import numpy as np

from ase.build import molecule
from gpaw import GPAW
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.poisson import PoissonSolver
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.tddft.spectrum import photoabsorption_spectrum
from gpaw.lrtddft2 import LrTDDFT2
from gpaw.mpi import world

from gpaw.test import equal

# Atoms


def test_lcaotddft_lcaotddft_vs_lrtddft2(in_tmp_dir):
    atoms = molecule('Na2')
    atoms.center(vacuum=4.0)

    # Ground-state calculation
    calc = GPAW(nbands=2, h=0.4, setups=dict(Na='1'),
                basis='sz(dzp)', mode='lcao', xc='oldLDA',
                poissonsolver=PoissonSolver(eps=1e-16),
                convergence={'density': 1e-8},
                txt='gs.out')
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')

    # Time-propagation calculation
    td_calc = LCAOTDDFT('gs.gpw', txt='td.out')
    DipoleMomentWriter(td_calc, 'dm.dat')
    td_calc.absorption_kick([0, 0, 1e-5])
    td_calc.propagate(30, 150)
    photoabsorption_spectrum('dm.dat', 'spec.dat',
                             e_max=10, width=0.5, delta_e=0.1)

    # LrTDDFT2 calculation
    calc = GPAW('gs.gpw', txt='lr.out')
    lr = LrTDDFT2('lr2', calc, fxc='LDA')
    lr.calculate()
    lr.get_spectrum('lr_spec.dat', 0, 10.1, 0.1, width=0.5)
    world.barrier()

    # Scale spectra due to slightly different definitions
    spec1_ej = np.loadtxt('spec.dat')
    data1_e = spec1_ej[:, 3]
    data1_e[1:] /= spec1_ej[1:, 0]
    spec2_ej = np.loadtxt('lr_spec.dat')
    E = lr.lr_transitions.get_transitions()[0][0]
    data2_e = spec2_ej[:, 5] / E

    # One can decrease the tolerance by decreasing the time step
    # and other parameters
    equal(data1_e, data2_e, 0.01)



from myqueue.task import task


def create_tasks():
    return [
        task('C_ecut_k_conv_GW.py@24:1d'),
        task('C_ecut_k_conv_plot_GW.py', deps='C_ecut_k_conv_GW.py'),
        task('C_ecut_extrap.py', deps='C_ecut_k_conv_GW.py'),
        task('C_frequency_conv.py@1:30m'),
        task('C_frequency_conv_plot.py', deps='C_frequency_conv.py'),
        task('C_equal_test.py',
             deps='C_ecut_k_conv_GW.py,C_frequency_conv.py'),
        task('BN_GW0.py@1:1h'),
        task('BN_GW0_plot.py', deps='BN_GW0.py'),
        task('MoS2_gs_GW.py@1:2h'),
        task('MoS2_GWG.py@8:20m', deps='MoS2_gs_GW.py'),
        task('MoS2_bs_plot.py', deps='MoS2_GWG.py'),
        task('check_gw.py', deps='MoS2_GWG.py')]



from myqueue.task import task


def create_tasks():
    return [task('run.py+16@16:10h'),
            task('run.py+8@8:12h'),
            task('run.py+4@4:5h'),
            task('run.py+1@1:1h'),
            task('analyse.py@1:10m',
                 deps=['run.py+16', 'run.py+8', 'run.py+4', 'run.py+1'])]



from gpaw import GPAW
from ase.io import read

L = 14  # ang
atoms = read('r-methyl-oxirane.xyz')
atoms.set_cell((L, L, L))
atoms.center()

calc = GPAW(h=0.25,
            nbands=30,
            convergence={'bands': 20},
            xc='LDA',
            maxiter=300,
            txt='unocc.txt')
atoms.calc = calc
atoms.get_potential_energy()

calc.write('r-methyl-oxirane.gpw', mode='all')



from ase.utils.sphinx import mol_role
from ase.utils.sphinx import git_role_tmpl
from ase.utils.sphinx import create_png_files


def git_role(role, rawtext, text, lineno, inliner, options={}, content=[]):
    return git_role_tmpl('https://gitlab.com/gpaw/gpaw/blob/master/',
                         role,
                         rawtext, text, lineno, inliner, options, content)

    
def setup(app):
    app.add_role('mol', mol_role)
    app.add_role('git', git_role)
    create_png_files()



""" Tests extrapolation to infinite energy cutoff + block paralellization.
It takes ~109 s on one core"""

import pytest
from gpaw.mpi import world
from gpaw.utilities import compiled_with_sl
import numpy as np

from ase.lattice.hexagonal import Graphene

from gpaw import GPAW, FermiDirac
from gpaw.test import equal
from gpaw.response.g0w0 import G0W0

pytestmark = pytest.mark.skipif(
    world.size != 1 and not compiled_with_sl(),
    reason='world.size != 1 and not compiled_with_sl()')


def test_response_gw_hBN_extrapolate(in_tmp_dir):
    if 1:
        calc = GPAW(mode='pw',
                    xc='PBE',
                    nbands=16,
                    setups={'Mo': '6'},
                    eigensolver='rmm-diis',
                    occupations=FermiDirac(0.001),
                    parallel={'domain': 1},
                    kpts={'size': (6, 6, 1), 'gamma': True})

        layer = Graphene(symbol='B',
                         latticeconstant={'a': 2.5, 'c': 1.0},
                         size=(1, 1, 1))
        layer[0].symbol = 'N'
        layer.pbc = (1, 1, 0)
        layer.center(axis=2, vacuum=4.0)

        layer.calc = calc
        layer.get_potential_energy()

        nbecut = 50
        from ase.units import Bohr, Hartree
        vol = layer.get_volume() / Bohr**3
        nbands = int(vol * (nbecut / Hartree)**1.5 * 2**0.5 / 3 / np.pi**2)
        calc.diagonalize_full_hamiltonian(nbands)
        calc.write('hBN.gpw', mode='all')

    gw = G0W0('hBN.gpw',
              'gw-hBN',
              ecut=50,
              domega0=0.1,
              eta=0.2,
              truncation='2D',
              kpts=[0],
              bands=(3, 5),
              ecut_extrapolation=[30, 40, 50],
              nblocksmax=True)

    e_qp = gw.calculate()['qp'][0, 0]

    ev = -4.381
    ec = 3.71013806
    equal(e_qp[0], ev, 0.01)
    equal(e_qp[1], ec, 0.01)



from distutils.version import LooseVersion
from gpaw import __ase_version_required__
from ase import __version__


def test_ase_features_ase3k_version():
    assert LooseVersion(__version__) >= __ase_version_required__



"""Test GW band-gaps for Si."""

import pytest
from gpaw.mpi import world
from gpaw.utilities import compiled_with_sl
from ase.build import bulk
import numpy as np

from gpaw.test import equal
from gpaw import GPAW, PW, FermiDirac
from gpaw.response.g0w0 import G0W0


pytestmark = pytest.mark.skipif(
    world.size != 1 and not compiled_with_sl(),
    reason='world.size != 1 and not compiled_with_sl()')


@pytest.mark.slow
def test_response_gwsi(in_tmp_dir):
    def run(atoms, symm, name):
        atoms.calc = GPAW(mode=PW(250),
                          eigensolver='rmm-diis',
                          occupations=FermiDirac(0.01),
                          symmetry=symm,
                          kpts={'size': (2, 2, 2), 'gamma': True},
                          parallel={'domain': 1},
                          txt=name + '.txt')
        e = atoms.get_potential_energy()
        scalapack = atoms.calc.wfs.bd.comm.size
        atoms.calc.diagonalize_full_hamiltonian(nbands=8, scalapack=scalapack)
        atoms.calc.write(name + '.gpw', mode='all')
        gw = G0W0(name + '.gpw', 'gw-' + name,
                  nbands=8,
                  integrate_gamma=0,
                  kpts=[(0, 0, 0), (0.5, 0.5, 0)],  # Gamma, X
                  ecut=40,
                  domega0=0.1,
                  eta=0.2,
                  relbands=(-1, 2))  # homo, lumo, lumo+1, same as bands=(3, 6)
        results = gw.calculate()
        return e, results

    a = 5.43
    si1 = bulk('Si', 'diamond', a=a)
    si2 = si1.copy()
    si2.positions -= a / 8

    i = 0
    results = []
    for si in [si1, si2]:
        for symm in [{},
                     'off',
                     {'time_reversal': False},
                     {'point_group': False}]:
            e, r = run(si, symm, str(i))
            G, X = r['eps'][0]
            results.append([e, G[0], G[1] - G[0], X[1] - G[0], X[2] - X[1]])
            G, X = r['qp'][0]
            results[-1].extend([G[0], G[1] - G[0], X[1] - G[0], X[2] - X[1]])
            i += 1

    equal(abs(np.array(results[0]) -
              [-9.25,
               5.44, 2.39, 0.40, 0,
               6.26, 3.57, 1.32, 0]).max(), 0, 0.025)
    equal(np.ptp(results, 0).max(), 0, 0.007)



import pickle
import pylab as plt
import numpy as np
from ase.units import kcal, mol

# VDWFunctional().make_prl_plot()

d = np.linspace(3.0, 5.5, 11)
S = ['Ar', 'Kr']
S = ['Kr']
for symbol in S:
    d, e, de = pickle.load(open(symbol + '.pckl', 'rb'))
    # plt.plot(d, e * 1000)
    plt.plot(d, (e + de) * 1000, label=symbol)
for symbol in S:
    e, de, e0, de0 = pickle.load(open(symbol + '.new.pckl', 'rb'))
    # plt.plot(d, (e-2*e0) * 1000)
    plt.plot(d, (e + de - 2 * e0 - 2 * de0) * 1000, label=symbol + '.new')
for symbol in S:
    e, de, e0, de0 = pickle.load(open(symbol + '.fft.pckl', 'rb'))
    plt.plot(d, (e - 2 * e0) * 1000)
    plt.plot(d, (e + de - 2 * e0 - 2 * de0) * 1000, label=symbol + '.fft')
plt.legend()
plt.ylim(-25, 50)
plt.show()


d, e, de = pickle.load(open('benzene.pckl', 'rb'))
plt.plot(d, e / (kcal / mol))
plt.plot(d, (e + de) / (kcal / mol))
e, de, e0, de0 = pickle.load(open('benzene.fft.pckl', 'rb'))
plt.plot(d, (e - 2 * e0) / (kcal / mol))
plt.plot(d, (e + de - 2 * e0 - 2 * de0) / (kcal / mol))
plt.xlim(3, 4.9)
plt.ylim(-4, 5)
plt.show()



# Creates: aluminum_EELS.png
import numpy as np
import matplotlib.pyplot as plt

data = np.loadtxt('eels.csv', delimiter=',')
plt.plot(data[:, 0], data[:, 2])
plt.savefig('aluminum_EELS.png', bbox_inches='tight')
plt.show()



import numpy as np
from gpaw.mpi import world
from parms import V, H, h1, ion_shift
from kgf.GreenFunctions import NonEqNonIntGreenFunction, NonEqIntGreenFunction
from kgf.ScatteringHamiltonianMatrix import ScatteringHamiltonianMatrix
from kgf.LeadHamiltonianMatrix import LeadHamiltonianMatrix
from kgf.selfenergies import NonEqConstantSelfEnergy
from kgf.selfenergies.GW2index import Hartree2index, Fock2index, GW2index

hmat = ScatteringHamiltonianMatrix(leftprincipallayer=1,
                                   rightprincipallayer=1,
                                   hamiltonianmatrix=H)

left = LeadHamiltonianMatrix(principallayer=1,
                             hamiltonianmatrix=h1)

right = LeadHamiltonianMatrix(principallayer=1,
                              hamiltonianmatrix=h1)

energies = np.arange(-96.0, 96.0, 0.02)
de = energies[1] - energies[0]
assert abs(energies).min() < 1.0e-8
fermi = 0.0
g0 = NonEqNonIntGreenFunction(hmat, left, right, E_Fermi=fermi,
                              energies=energies)
g0.SetInfinitesimal(2 * de)
Ntot = g0.GetTotalParticleNumber()
if world.rank == 0:
    print('Total electron number:', Ntot)
se_null = NonEqConstantSelfEnergy(g0, np.zeros((6, 6), complex), 'null')
se_ion = NonEqConstantSelfEnergy(g0, ion_shift, 'shift')
se_hartree = Hartree2index(g0, V, initialhartree=np.zeros((6, 6), complex))
se_fock = Fock2index(g0, V)
se_gw = GW2index(g0, V)
gf = NonEqIntGreenFunction([se_null, se_ion, se_hartree, se_fock, se_gw])
orbitals = range(1, 5)
pulay = (0.03, 0.25, 1)
# Non interacting calculation
se_ion.SetActive(False)
se_hartree.SetActive(False)
se_fock.SetActive(False)
se_gw.SetActive(False)
gf.WriteSpectralInfoToNetCDFFile('nonint.nc',
                                 diagonalize=True,
                                 orbitals=orbitals,
                                 spectral='individual')
# HF calculation
se_ion.SetActive(True)
se_hartree.SetActive(True)
se_fock.SetActive(True)
gf.SelfConsistent(log='HF.log', pulay=pulay)
gf.WriteSpectralInfoToNetCDFFile('HF.nc',
                                 diagonalize=True,
                                 orbitals=orbitals,
                                 spectral='individual')

# GW calculation
se_gw.SetActive(True)
gf.SelfConsistent(log='GW.log', pulay=pulay)
gf.WriteSpectralInfoToNetCDFFile('GW.nc',
                                 diagonalize=True,
                                 orbitals=orbitals,
                                 spectral='individual')



from ase.build import bulk
from gpaw import GPAW, FermiDirac
from gpaw.wavefunctions.pw import PW

a = 3.567
atoms = bulk('C', 'diamond', a=a)

calc = GPAW(mode=PW(300),  # energy cutoff for plane wave basis (in eV)
            kpts={'size': (3, 3, 3), 'gamma': True},
            xc='LDA',
            occupations=FermiDirac(0.001),
            parallel={'domain': 1},
            txt='C_groundstate.txt')

atoms.calc = calc
atoms.get_potential_energy()

calc.diagonalize_full_hamiltonian()  # determine all bands
calc.write('C_groundstate.gpw', 'all')  # write out wavefunctions



# creates: bigpicture.svg, bigpicture.png
from math import pi, cos, sin

import numpy as np
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt


class Box:
    def __init__(self, name, description=(), attributes=(), color='grey'):
        self.name = name
        if isinstance(description, str):
            description = [description]
        self.description = description
        self.attributes = attributes
        self.color = color

        self.owns = []
        self.position = None

    def set_position(self, position):
        self.position = np.asarray(position)

    def has(self, other, name, angle=None, distance=None, x=0.4, style='<-'):
        self.owns.append((other, name, x, style))
        if angle is not None:
            angle *= pi / 180
            other.set_position(self.position +
                               [cos(angle) * distance, sin(angle) * distance])


def cut(size, dir):
    if abs(size[0] * dir[1]) < abs(size[1] * dir[0]):
        x = min(max(-size[0] / 2, dir[0]), size[0] / 2)
        y = x * dir[1] / dir[0]
    else:
        y = min(max(-size[1] / 2, dir[1]), size[1] / 2)
        x = y * dir[0] / dir[1]
    return x, y


class MPL:
    def __init__(self, boxes):
        self.boxes = boxes

    def plot(self):
        a4 = 100 * np.array([2**-1.75, 2**-2.25])
        inch = 2.54

        self.fig = plt.figure(1, a4 / inch)
        self.ax = ax = self.fig.add_axes([0, 0, 1, 1], frameon=False)
        ax.set_xlim(0, a4[0])
        ax.set_ylim(0, a4[1])
        ax.xaxis.set_visible(False)
        ax.yaxis.set_visible(False)

        ax.add_patch(mpatches.Rectangle((22.5, 16), 6, 4, fc='orange'))
        ax.text(22.7, 19.5, 'ASE package')

        for b in boxes:
            x, y = b.position

            text = b.name
            for txt in b.description:
                text += '\n' + txt
            for txt in b.attributes:
                text += '\n' + txt

            b.text = ax.text(x, y,
                             text,
                             fontsize=9,
                             ha='center',
                             va='center',
                             bbox=dict(boxstyle='round',
                                       facecolor=b.color,
                                       alpha=0.75))

        self.fig.canvas.mpl_connect('draw_event', self.on_draw)
        plt.savefig('bigpicture.png', dpi=50)
        plt.savefig('bigpicture.svg')

    def on_draw(self, event):
        for b in self.boxes:
            bbox = b.text.get_window_extent()
            t = b.text.get_transform()
            b.size = t.inverted().transform(bbox.size)

        for b in self.boxes:
            for other, name, s, style in b.owns:
                d = other.position - b.position
                p1 = b.position + cut(b.size, d)
                p2 = other.position + cut(other.size, -d)
                if style == '-|>':
                    arrowprops = dict(arrowstyle=style, fc='white')
                else:
                    arrowprops = dict(arrowstyle=style)

                self.ax.annotate('', p1, p2,
                                 arrowprops=arrowprops)
                if name:
                    p = (1 - s) * p1 + s * p2
                    self.ax.text(p[0], p[1], name, fontsize=7,
                                 ha='center', va='center',
                                 bbox=dict(facecolor='white', ec='white'))

        self.fig.canvas.callbacks.callbacks[event.name] = {}
        self.fig.canvas.draw()
        return False


boxes = []


def box(*args, **kwargs):
    b = Box(*args, **kwargs)
    boxes.append(b)
    return b


atoms = box('Atoms', [''], ['positions, numbers, cell, pbc'],
            color='white')
paw = box('PAW', [], [], 'green')
scf = box('SCFLoop', [])
density = box('Density',
              [r'$\tilde{n}_\sigma = \sum_{\mathbf{k}n}' +
               r'|\tilde{\psi}_{\sigma\mathbf{k}n}|^2' +
               r'+\frac{1}{2}\sum_a \tilde{n}_c^a$',
               r'$\tilde{\rho}(\mathbf{r}) = ' +
               r'\sum_\sigma\tilde{n}_\sigma + \sum_{aL}Q_L^a \hat{g}_L^a$'],
              ['nspins, nt_sG, nt_sg,', 'rhot_g, Q_aL, D_asp'])
mixer = box('Mixer')
hamiltonian = box('Hamiltonian',
                  [r'$-\frac{1}{2}\nabla^2 + \tilde{v} + ' +
                   r'\sum_a \sum_{i_1i_2} |\tilde{p}_{i_1}^a \rangle ' +
                   r'\Delta H_{i_1i_2} \langle \tilde{p}_{i_2}^a|$'],
                  ['nspins, vt_sG, vt_sg, vHt_g, dH_asp',
                   'Etot, Ekin, Exc, e_coulomb, Ebar'])
wfs = box('WaveFunctions',
          [r'$\tilde{\psi}_{\sigma\mathbf{k}n}(\mathbf{r})$'],
          ['nspins, ibzk_qc, mynbands',
           'kpt_comm, band_comm'], color='magenta')
gd = box('GridDescriptor', ['(coarse grid)'],
         ['cell_cv, N_c,', 'pbc_c, dv, comm'], 'orange')
finegd = box('GridDescriptor', '(fine grid)',
             ['cell_cv, N_c, pbc_c, dv, comm'], 'orange')
rgd = box('RadialGridDescriptor', [], ['r_g, dr_g, rcut'], color='orange')
setups = box('Setups', ['', '', '', ''], ['nvalence, nao, Eref, corecharge'])
xccorrection = box('XCCorrection')
nct = box('LFC', r'$\tilde{n}_c^a(r)$', [], 'red')
vbar = box('LFC', r'$\bar{v}^a(r)$', [], 'red')
ghat = box('LFC', r'$\hat{g}_{\ell m}^a(\mathbf{r})$', [], 'red')
fd = box('FDWaveFunctions',
         r"""$\tilde{\psi}_{\sigma\mathbf{k}n}(ih,jh,kh)$""",
         [], 'magenta')
pt = box('LFC', r'$\tilde{p}_i^a(\mathbf{r})$', [], 'red')
lcao = box('LCAOWaveFunctions',
           r"$\tilde{\psi}_{\sigma\mathbf{k}n}(\mathbf{r})=" +
           r"\sum_{\mu\mathbf{R}} " +
           r"C_{\sigma\mathbf{k}n\mu} \Phi_\mu(\mathbf{r} - \mathbf{R}) " +
           r"\exp(i\mathbf{k}\cdot\mathbf{R})$",
           ['S_qMM, T_qMM, P_aqMi'], 'magenta')
atoms0 = box('Atoms', '(copy)', ['positions, numbers, cell, pbc'],
             color='grey')
parameters = box('InputParameters', [], ['xc, nbands, ...'])
forces = box('ForceCalculator')
occupations = box(
    'OccupationNumbers',
    r'$\epsilon_{\sigma\mathbf{k}n} \rightarrow f_{\sigma\mathbf{k}n}$')
poisson = box(
    'PoissonSolver',
    r'$\nabla^2 \tilde{v}_H(\mathbf{r}) = -4\pi \tilde{\rho}(\mathbf{r})$')
eigensolver = box('EigenSolver')
symmetry = box('Symmetry')
restrictor = box('Transformer', '(fine -> coarse)',
                 color='yellow')
interpolator = box('Transformer', '(coarse -> fine)',
                   color='yellow')
xc = box('XCFunctional')
kin = box('FDOperator', r'$-\frac{1}{2}\nabla^2$')
hsoperator = box('HSOperator',
                 [r"$\langle \psi_n | A | \psi_{n'} \rangle$",
                  r"$\sum_{n'}U_{nn'}|\tilde{\psi}_{n'}\rangle$"])

overlap = box('Overlap')
basisfunctions = box('BasisFunctions', r'$\Phi_\mu(\mathbf{r})$',
                     color='red')
tci = box('TwoCenterIntegrals',
          r'$\langle\Phi_\mu|\Phi_\nu\rangle,'
          r'\langle\Phi_\mu|\hat{T}|\Phi_\nu\rangle,'
          r'\langle\tilde{p}^a_i|\Phi_\mu\rangle$')

atoms.set_position((25, 18.3))
atoms.has(paw, 'calculator', -160, 7.5)
paw.has(scf, 'scf', 160, 4, x=0.48)
paw.has(density, 'density', -150, 14, 0.23)
paw.has(hamiltonian, 'hamiltonian', 180, 10, 0.3)
paw.has(wfs, 'wfs', -65, 5.5, x=0.48)
paw.has(atoms0, 'atoms', 9, 7.5)
paw.has(parameters, 'input_parameters', 90, 4)
paw.has(forces, 'forces', 50, 4)
paw.has(occupations, 'occupations', 136, 4)
density.has(mixer, 'mixer', 130, 3.3)
density.has(gd, 'gd', x=0.33)
density.has(finegd, 'finegd', 76, 3.5)
density.has(setups, 'setups', 0, 7, 0.45)
density.has(nct, 'nct', -90, 3)
density.has(ghat, 'ghat', -130, 3.4)
density.has(interpolator, 'interpolator', -45, 4)
hamiltonian.has(restrictor, 'restrictor', 40, 4)
hamiltonian.has(xc, 'xc', 160, 6, x=0.6)
hamiltonian.has(vbar, 'vbar', 80, 4)
hamiltonian.has(setups, 'setups', x=0.3)
hamiltonian.has(gd, 'gd', x=0.45)
hamiltonian.has(finegd, 'finegd')
hamiltonian.has(poisson, 'poissonsolver', 130, 4)
wfs.has(gd, 'gd', 160, 4.8, x=0.48)
wfs.has(setups, 'setups', x=0.4)
wfs.has(lcao, None, -55, 5.9, style='-|>')
wfs.has(fd, None, -112, 5.0, style='-|>')
wfs.has(eigensolver, 'eigensolver', 30, 5, x=0.6)
wfs.has(symmetry, 'symmetry', 80, 3)
fd.has(pt, 'pt', -45, 3.6)
fd.has(kin, 'kin', -90, 3)
fd.has(overlap, 'overlap', -135, 3.5)
lcao.has(basisfunctions, 'basis_functions', -50, 3.5)
lcao.has(tci, 'tci', -90, 4.2)
overlap.has(setups, 'setups', x=0.4)
overlap.has(hsoperator, 'operator', -115, 2.5, x=0.41)

for i in range(3):
    setup = box('Setup', [],
                ['Z, Nv, Nc, pt_j, nct,', 'vbar, ghat_l, Delta_pl'],
                'blue')
    setup.set_position(setups.position +
                       (0.9 - i * 0.14, 0.3 - i * 0.14))
setup.has(xccorrection, 'xc_correction', -110, 3.7)
xccorrection.has(rgd, 'rgd', -105, 2.4, 0.4)

kpts = [box('KPoint', [], ['psit_nG, C_nM,', 'eps_n, f_n, P_ani'],
            color='cyan') for i in range(3)]
wfs.has(kpts[1], 'kpt_u', 0, 5.4, 0.48)
kpts[0].set_position(kpts[1].position - 0.14)
kpts[2].set_position(kpts[1].position + 0.14)

MPL(boxes).plot()



from myqueue.task import task


def create_tasks():
    t1 = task('cu_calc.py', cores=4, tmax='1h')
    t2 = task('cu_plot.py', deps=t1, creates='cu.png')
    t3 = task('cu_agts.py', deps=t1)
    return [t1, t2, t3]


def check():
    import numpy as np
    from ase.io import read
    energies = []
    k = 20
    for name in ['ITM', 'FD-0.05', 'MV-0.2']:
        e = read(f'Cu-{name}-{k}.txt').get_potential_energy()
        energies.append(e)
    # Extrapolate TM:
    e19 = read('Cu-TM-19.txt').get_potential_energy()
    e20 = read('Cu-TM-20.txt').get_potential_energy()
    e = np.polyval(np.polyfit([20**-2, 19**-2], [e20, e19], 1), 0)
    energies.append(e)
    assert max(energies) - min(energies) < 0.001


if __name__ == '__main__':
    check()



# %%
"""
# Discover a new 2D material with high critical temperature

Now that you know how to calculate exchange coupling constants and anisotropies, you are in a position to search for new magnetic materials yourself.

Go to the 2D database at https://cmrdb.fysik.dtu.dk/?project=c2db and search for ferromagnetic magnetic (FM) or anti-ferromagnetic (AFM) materials with finite band gaps. (The theory developed so far is not applicable to metals, since these typically have long range interactions). Click the Add Column button on the right and add the magnetic anisotropy. You should now be able to sort the 2D materials according to anisotropy and choose a material with a large value. A large anisotropy is expected to lead to a high critical temperature. Large values of $J$ and $S$ are also desirable and $J$ may be estimated from the deviation from the paramagnetic state stated at the webpage.

Decide on a material you find promising and perform the analysis from the previous notebooks. I suggest you do not choose an anti-ferromagnetic hexagonal material, since the non-collinear ground state makes the analysis somewhat complicated. Calculate $J$ and $A$ and estimate the critical temperature from the expression stated in part 1. As a rough estimate you may also use the expression for anti-ferromagnetic systems if that is what you decide on.
"""

# %%



def create_tasks():
    from myqueue.task import task
    return [
        task('h2o.py', folder='water'),
        task('CO.py@8:15m', folder='wavefunctions'),
        task('Al_fcc.py@2:15m', folder='aluminium'),
        task('Al_bcc.py@2:15m', folder='aluminium'),
        task('Al_fcc_vs_bcc.py@2:15s', folder='aluminium'),
        task('Al_fcc_modified.py@2:15s', folder='aluminium'),
        task('initial.py@2:15m', folder='diffusion'),
        task('solution.py@2:15m', folder='diffusion'),
        task('densitydiff.py', folder='diffusion', deps='solution.py'),
        task('si.py@8:15m', folder='wannier'),
        task('wannier-si.py', folder='wannier', deps='si.py'),
        task('benzene.py@8:15m', folder='wannier'),
        task('wannier-benzene.py', folder='wannier', deps='benzene.py'),
        task('ag.py', folder='band_structure'),
        task('h2o.py@8:15m', folder='vibrations'),
        task('H2O_vib.py@8:15m', folder='vibrations', deps='h2o.py'),
        task('H2O_vib_2.py@4:15m', folder='vibrations', deps='H2O_vib.py'),
        task('ferro.py@4:15m', folder='iron'),
        task('anti.py@4:15m', folder='iron'),
        task('non.py@2:15m', folder='iron'),
        task('PBE.py', folder='iron', deps='ferro.py,anti.py,non.py'),
        task('test.py', folder='eels', deps='../band_structure/ag.py'),
        task('test.py', folder='gw'),
        task('si.pbe.py', folder='rpa'),
        task('si_pbe_exx.py@4:15m', folder='rpa', deps='si.pbe.py'),
        task('si.rpa_init_pbe.py', folder='rpa'),
        task('si.rpa.py@4:15m', folder='rpa', deps='si.rpa_init_pbe.py'),
        task('con_pw.py', folder='stress'),
        task('stress.py', folder='stress')
        # task('pt_h2_tb_transport.py', folder='transport'),
        # task('pt_h2_lcao_manual.py', folder='transport'),
        # task('pt_h2_lcao_transport.py', folder='transport',
        #      deps='pt_h2_lcao_manual.py')
    ]




from math import sqrt
import numpy as np
from ase import Atoms
from ase.parallel import paropen
from gpaw import GPAW, FermiDirac, PW
from gpaw.response.df import DielectricFunction

# Part 1: Ground state calculation
a = 1.42
c = 3.355

# Generate graphite AB-stack structure:
atoms = Atoms('C4',
              scaled_positions=[(1 / 3.0, 1 / 3.0, 0),
                                (2 / 3.0, 2 / 3.0, 0),
                                (0, 0, 0.5),
                                (1 / 3.0, 1 / 3.0, 0.5)],
              pbc=(1, 1, 1),
              cell=[(sqrt(3) * a / 2, 3 / 2.0 * a, 0),
                    (-sqrt(3) * a / 2, 3 / 2.0 * a, 0),
                    (0, 0, 2 * c)])

# Part 2: Find ground state density and diagonalize full hamiltonian
calc = GPAW(mode=PW(500),
            kpts=(6, 6, 3),
            parallel={'domain': 1},
            # Use smaller Fermi-Dirac smearing to avoid intraband transitions:
            occupations=FermiDirac(0.05))

atoms.calc = calc
atoms.get_potential_energy()

calc = calc.fixed_density(kpts=(20, 20, 7))

# The result should also be converged with respect to bands:
calc.diagonalize_full_hamiltonian(nbands=60)
calc.write('graphite.gpw', 'all')

# Part 2: Spectra calculations
f = paropen('graphite_q_list', 'w')  # write q

for i in range(1, 6):  # loop over different q
    df = DielectricFunction(calc='graphite.gpw',
                            domega0=0.01,
                            eta=0.2,  # Broadening parameter.
                            ecut=100,
                            # write different output for different q:
                            txt='out_df_%d.txt' % i)

    q_c = [i / 20.0, 0.0, 0.0]  # Gamma - M excitation

    df.get_eels_spectrum(q_c=q_c, filename='graphite_EELS_%d' % i)

    # Calculate cartesian momentum vector:
    cell_cv = atoms.get_cell()
    bcell_cv = 2 * np.pi * np.linalg.inv(cell_cv).T
    q_v = np.dot(q_c, bcell_cv)
    print(sqrt(np.inner(q_v, q_v)), file=f)

f.close()



"""This module implements a class for (true) time-dependent density
functional theory calculations.

"""
import time
from math import log

import numpy as np

from gpaw.calculator import GPAW
from gpaw.mixer import DummyMixer
from gpaw.preconditioner import Preconditioner
from gpaw.tddft.units import (attosec_to_autime, autime_to_attosec,
                              aufrequency_to_eV)
from gpaw.tddft.utils import MultiBlas
from gpaw.tddft.bicgstab import BiCGStab
from gpaw.tddft.cscg import CSCG
from gpaw.tddft.propagators import \
    ExplicitCrankNicolson, \
    SemiImplicitCrankNicolson, \
    EhrenfestPAWSICN,\
    EhrenfestHGHSICN,\
    EnforcedTimeReversalSymmetryCrankNicolson, \
    SemiImplicitTaylorExponential, \
    SemiImplicitKrylovExponential, \
    AbsorptionKick
from gpaw.tddft.tdopers import \
    TimeDependentHamiltonian, \
    TimeDependentOverlap, \
    TimeDependentWaveFunctions, \
    TimeDependentDensity, \
    AbsorptionKickHamiltonian
from gpaw.wavefunctions.fd import FD

from gpaw.tddft.spectrum import photoabsorption_spectrum

photoabsorption_spectrum = photoabsorption_spectrum


# T^-1
# Bad preconditioner
class KineticEnergyPreconditioner:
    def __init__(self, gd, kin, dtype):
        self.preconditioner = Preconditioner(gd, kin, dtype)
        self.preconditioner.allocate()

    def apply(self, kpt, psi, psin):
        for i in range(len(psi)):
            psin[i][:] = self.preconditioner(psi[i], kpt.phase_cd, None, None)


# S^-1
class InverseOverlapPreconditioner:
    """Preconditioner for TDDFT."""
    def __init__(self, overlap):
        self.overlap = overlap

    def apply(self, kpt, psi, psin):
        self.overlap.apply_inverse(psi, psin, kpt)
# ^^^^^^^^^^


class FDTDDFTMode(FD):
    def __call__(self, *args, **kwargs):
        reuse_wfs_method = kwargs.pop('reuse_wfs_method', None)
        assert reuse_wfs_method is None
        return TimeDependentWaveFunctions(self.nn, *args, **kwargs)


class TDDFT(GPAW):
    """Time-dependent density functional theory calculation based on GPAW.

    This class is the core class of the time-dependent density functional
    theory implementation and is the only class which a user has to use.
    """

    def __init__(self, filename,
                 td_potential=None, propagator='SICN', calculate_energy=True,
                 propagator_kwargs=None, solver='CSCG', tolerance=1e-8,
                 **kwargs):
        """Create TDDFT-object.

        Parameters:

        filename: string
            File containing ground state or time-dependent state to propagate
        td_potential: class, optional
            Function class for the time-dependent potential. Must have a method
            'strength(time)' which returns the strength of the linear potential
            to each direction as a vector of three floats.
        propagator:  {'SICN','ETRSCN','ECN','SITE','SIKE4','SIKE5','SIKE6'}
            Name of the time propagator for the Kohn-Sham wavefunctions
        solver: {'CSCG','BiCGStab'}
            Name of the iterative linear equations solver for time propagation
        tolerance: float
            Tolerance for the linear solver

        The following parameters can be used: `txt`, `parallel`, `communicator`
        `mixer` and `dtype`. The internal parameters `mixer` and `dtype` are
        strictly used to specify a dummy mixer and complex type respectively.
        """

        # Set initial time
        self.time = 0.0

        # Set initial kick strength
        self.kick_strength = np.array([0.0, 0.0, 0.0], dtype=float)

        # Set initial value of iteration counter
        self.niter = 0

        # Parallelization dictionary should default to strided bands
        self.default_parallel = GPAW.default_parallel.copy()
        self.default_parallel['stridebands'] = True

        self.default_parameters = GPAW.default_parameters.copy()
        self.default_parameters['mixer'] = DummyMixer()
        self.default_parameters['symmetry'] = {'point_group': False}

        # NB: TDDFT restart files contain additional information which
        #     will override the initial settings for time/kick/niter.
        GPAW.__init__(self, filename, **kwargs)

        assert isinstance(self.wfs, TimeDependentWaveFunctions)
        assert isinstance(self.wfs.overlap, TimeDependentOverlap)

        # Prepare for dipole moment file handle
        self.dm_file = None

        # Initialize wavefunctions and density
        # (necessary after restarting from file)
        if not self.initialized:
            self.initialize()
        self.set_positions()

        # Don't be too strict
        self.density.charge_eps = 1e-5

        wfs = self.wfs
        self.rank = wfs.world.rank

        self.text = self.log
        self.text('')
        self.text('')
        self.text('------------------------------------------')
        self.text('  Time-propagation TDDFT                  ')
        self.text('------------------------------------------')
        self.text('')

        self.text('Charge epsilon: ', self.density.charge_eps)

        # Time-dependent variables and operators
        self.td_potential = td_potential
        self.td_hamiltonian = TimeDependentHamiltonian(self.wfs, self.spos_ac,
                                                       self.hamiltonian,
                                                       td_potential)
        self.td_overlap = self.wfs.overlap  # TODO remove this property
        self.td_density = TimeDependentDensity(self)

        # Solver for linear equations
        self.text('Solver: ', solver)
        if solver == 'BiCGStab':
            self.solver = BiCGStab(gd=wfs.gd, timer=self.timer,
                                   tolerance=tolerance)
        elif solver == 'CSCG':
            self.solver = CSCG(gd=wfs.gd, timer=self.timer,
                               tolerance=tolerance)
        else:
            raise RuntimeError('Solver %s not supported.' % solver)

        # Preconditioner
        # No preconditioner as none good found
        self.text('Preconditioner: ', 'None')
        self.preconditioner = None  # TODO! check out SSOR preconditioning
        # self.preconditioner = InverseOverlapPreconditioner(self.overlap)
        # self.preconditioner = KineticEnergyPreconditioner(
        #    wfs.gd, self.td_hamiltonian.hamiltonian.kin, complex)

        # Time propagator
        self.text('Propagator: ', propagator)
        if propagator_kwargs is None:
            propagator_kwargs = {}
        if propagator == 'ECN':
            self.propagator = ExplicitCrankNicolson(
                self.td_density,
                self.td_hamiltonian, self.td_overlap, self.solver,
                self.preconditioner, wfs.gd, self.timer, **propagator_kwargs)
        elif propagator == 'SICN':
            self.propagator = SemiImplicitCrankNicolson(
                self.td_density,
                self.td_hamiltonian, self.td_overlap, self.solver,
                self.preconditioner, wfs.gd, self.timer, **propagator_kwargs)
        elif propagator == 'EFSICN':
            self.propagator = EhrenfestPAWSICN(
                self.td_density,
                self.td_hamiltonian, self.td_overlap, self.solver,
                self.preconditioner, wfs.gd, self.timer, **propagator_kwargs)
        elif propagator == 'EFSICN_HGH':
            self.propagator = EhrenfestHGHSICN(
                self.td_density,
                self.td_hamiltonian, self.td_overlap, self.solver,
                self.preconditioner, wfs.gd, self.timer, **propagator_kwargs)
        elif propagator == 'ETRSCN':
            self.propagator = EnforcedTimeReversalSymmetryCrankNicolson(
                self.td_density,
                self.td_hamiltonian, self.td_overlap, self.solver,
                self.preconditioner, wfs.gd, self.timer, **propagator_kwargs)
        elif propagator == 'SITE':
            self.propagator = SemiImplicitTaylorExponential(
                self.td_density,
                self.td_hamiltonian, self.td_overlap, self.solver,
                self.preconditioner, wfs.gd, self.timer, **propagator_kwargs)
        elif propagator == 'SIKE':
            self.propagator = SemiImplicitKrylovExponential(
                self.td_density,
                self.td_hamiltonian, self.td_overlap, self.solver,
                self.preconditioner, wfs.gd, self.timer, **propagator_kwargs)
        elif propagator.startswith('SITE') or propagator.startswith('SIKE'):
            raise DeprecationWarning(
                'Use propagator_kwargs to specify degree.')
        else:
            raise RuntimeError(
                'Time propagator %s not supported.' % propagator)

        if self.rank == 0:
            if wfs.kd.comm.size > 1:
                if wfs.nspins == 2:
                    self.text('Parallelization Over Spin')

                if wfs.gd.comm.size > 1:
                    self.text('Using Domain Decomposition: %d x %d x %d' %
                              tuple(wfs.gd.parsize_c))

                if wfs.bd.comm.size > 1:
                    self.text('Parallelization Over bands on %d Processors' %
                              wfs.bd.comm.size)
            self.text('States per processor = ', wfs.bd.mynbands)

        self.hpsit = None
        self.eps_tmp = None
        self.mblas = MultiBlas(wfs.gd)

        # Restarting an FDTD run generates hamiltonian.fdtd_poisson, which
        # now overwrites hamiltonian.poisson
        if hasattr(self.hamiltonian, 'fdtd_poisson'):
            self.hamiltonian.poisson = self.hamiltonian.fdtd_poisson
            self.hamiltonian.poisson.set_grid_descriptor(self.density.finegd)

        # For electrodynamics mode
        if self.hamiltonian.poisson.get_description() == 'FDTD+TDDFT':
            self.initialize_FDTD()
            self.hamiltonian.poisson.print_messages(self.text)
            self.log.flush()

        self.calculate_energy = calculate_energy
        if self.hamiltonian.xc.name.startswith('GLLB'):
            self.text('GLLB model potential. Not updating energy.')
            self.calculate_energy = False

    def create_wave_functions(self, mode, *args, **kwargs):
        mode = FDTDDFTMode(mode.nn, mode.interpolation, True)
        GPAW.create_wave_functions(self, mode, *args, **kwargs)

    def read(self, filename):
        reader = GPAW.read(self, filename)
        if 'tddft' in reader:
            self.time = reader.tddft.time
            self.niter = reader.tddft.niter
            self.kick_strength = reader.tddft.kick_strength

    def initialize(self, reading=False):
        self.parameters.mixer = DummyMixer()
        self.parameters.experimental['reuse_wfs_method'] = None
        GPAW.initialize(self, reading=reading)

    def _write(self, writer, mode):
        GPAW._write(self, writer, mode)
        writer.child('tddft').write(time=self.time,
                                    niter=self.niter,
                                    kick_strength=self.kick_strength)

    # Electrodynamics requires extra care
    def initialize_FDTD(self):

        # Sanity check
        assert(self.hamiltonian.poisson.get_description() == 'FDTD+TDDFT')

        self.hamiltonian.poisson.set_density(self.density)

        # The propagate calculation_mode causes classical part to evolve
        # in time when self.hamiltonian.poisson.solve(...) is called
        self.hamiltonian.poisson.set_calculation_mode('propagate')

        # During each time step, self.hamiltonian.poisson.solve may be called
        # several times (depending on the used propagator). Using the
        # attached observer one ensures that actual propagation takes
        # place only once. This is because
        # the FDTDPoissonSolver changes the calculation_mode from propagate to
        # something else when the propagation is finished.
        self.attach(self.hamiltonian.poisson.set_calculation_mode, 1,
                    'propagate')

    def propagate(self, time_step, iterations, dipole_moment_file=None,
                  restart_file=None, dump_interval=100):
        """Propagates wavefunctions.

        Parameters:

        time_step: float
            Time step in attoseconds (10^-18 s), e.g., 4.0 or 8.0
        iterations: integer
            Iterations, e.g., 20 000 as / 4.0 as = 5000
        dipole_moment_file: string, optional
            Name of the data file where to the time-dependent dipole
            moment is saved
        restart_file: string, optional
            Name of the restart file
        dump_interval: integer
            After how many iterations restart data is dumped

        """

        if self.rank == 0:
            self.text()
            self.text('Starting time: %7.2f as'
                      % (self.time * autime_to_attosec))
            self.text('Time step:     %7.2f as' % time_step)
            header = """\
                        Simulation     Total         log10     Iterations:
             Time          time        Energy (eV)   Norm      Propagator"""
            self.text()
            self.text(header)

        # Convert to atomic units
        time_step = time_step * attosec_to_autime

        if dipole_moment_file is not None:
            self.initialize_dipole_moment_file(dipole_moment_file)

        # Set these as class properties for use of observers
        self.time_step = time_step
        self.dump_interval = dump_interval

        niterpropagator = 0
        self.maxiter = self.niter + iterations

        # Let FDTD part know the time step
        if self.hamiltonian.poisson.get_description() == 'FDTD+TDDFT':
            self.hamiltonian.poisson.set_time(self.time)
            self.hamiltonian.poisson.set_time_step(self.time_step)

        self.timer.start('Propagate')
        while self.niter < self.maxiter:
            norm = self.density.finegd.integrate(self.density.rhot_g)

            # Write dipole moment at every iteration
            if dipole_moment_file is not None:
                self.update_dipole_moment_file(norm)

            # print output (energy etc.) every 10th iteration
            if self.niter % 10 == 0:
                self.get_td_energy()

                T = time.localtime()
                if self.rank == 0:
                    iter_text = 'iter: %3d  %02d:%02d:%02d %11.2f' \
                                '   %13.6f %9.1f %10d'
                    self.text(iter_text %
                              (self.niter, T[3], T[4], T[5],
                               self.time * autime_to_attosec,
                               self.Etot * aufrequency_to_eV,
                               log(abs(norm) + 1e-16) / log(10),
                               niterpropagator))

                    self.log.flush()

            # Propagate the Kohn-Shame wavefunctions a single timestep
            niterpropagator = self.propagator.propagate(self.time, time_step)
            self.time += time_step

            # Call registered callback functions
            self.call_observers(self.niter)
            self.niter += 1

            # Write restart data
            if restart_file is not None and self.niter % dump_interval == 0:
                self.write(restart_file, 'all')
                if self.rank == 0:
                    print('Wrote restart file.')
                    print(self.niter, ' iterations done. Current time is ',
                          self.time * autime_to_attosec, ' as.')

        self.timer.stop('Propagate')

        # Write final results and close dipole moment file
        if dipole_moment_file is not None:
            # TODO final iteration is propagated, but nothing is updated
            # norm = self.density.finegd.integrate(self.density.rhot_g)
            # self.finalize_dipole_moment_file(norm)
            self.finalize_dipole_moment_file()

        # Finalize FDTDPoissonSolver
        if self.hamiltonian.poisson.get_description() == 'FDTD+TDDFT':
            self.hamiltonian.poisson.finalize_propagation()

        # Call registered callback functions
        self.call_observers(self.niter, final=True)

        if restart_file is not None:
            self.write(restart_file, 'all')

    def initialize_dipole_moment_file(self, dipole_moment_file):
        if self.rank == 0:
            if self.dm_file is not None and not self.dm_file.closed:
                raise RuntimeError('Dipole moment file is already open')

            if self.time == 0.0:
                mode = 'w'
            else:
                # We probably continue from restart
                mode = 'a'

            self.dm_file = open(dipole_moment_file, mode)

            # If the dipole moment file is empty, add a header
            if self.dm_file.tell() == 0:
                header = '# Kick = [%22.12le, %22.12le, %22.12le]\n' \
                    % (self.kick_strength[0], self.kick_strength[1],
                       self.kick_strength[2])
                header += '# %15s %15s %22s %22s %22s\n' \
                    % ('time', 'norm', 'dmx', 'dmy', 'dmz')
                self.dm_file.write(header)
                self.dm_file.flush()

    def update_dipole_moment_file(self, norm):
        dm = self.density.finegd.calculate_dipole_moment(self.density.rhot_g)

        if self.hamiltonian.poisson.get_description() == 'FDTD+TDDFT':
            dm += self.hamiltonian.poisson.get_classical_dipole_moment()

        if self.rank == 0:
            line = '%20.8lf %20.8le %22.12le %22.12le %22.12le\n' \
                % (self.time, norm, dm[0], dm[1], dm[2])
            self.dm_file.write(line)
            self.dm_file.flush()

    def finalize_dipole_moment_file(self, norm=None):
        if norm is not None:
            self.update_dipole_moment_file(norm)

        if self.rank == 0:
            self.dm_file.close()
            self.dm_file = None

    def update_eigenvalues(self):

        kpt_u = self.wfs.kpt_u
        if self.hpsit is None:
            self.hpsit = self.wfs.gd.zeros(len(kpt_u[0].psit_nG),
                                           dtype=complex)
        if self.eps_tmp is None:
            self.eps_tmp = np.zeros(len(kpt_u[0].eps_n),
                                    dtype=complex)

        # self.Eband = sum_i <psi_i|H|psi_j>
        for kpt in kpt_u:
            self.td_hamiltonian.apply(kpt, kpt.psit_nG, self.hpsit,
                                      calculate_P_ani=False)
            self.mblas.multi_zdotc(self.eps_tmp, kpt.psit_nG,
                                   self.hpsit, len(kpt_u[0].psit_nG))
            self.eps_tmp *= self.wfs.gd.dv
            kpt.eps_n[:] = self.eps_tmp.real

        H = self.td_hamiltonian.hamiltonian

        # Nonlocal
        self.Enlkin = H.xc.get_kinetic_energy_correction()

        # PAW
        e_band = self.wfs.calculate_band_energy()
        self.Ekin = H.e_kinetic0 + e_band + self.Enlkin
        self.e_coulomb = H.e_coulomb
        self.Eext = H.e_external
        self.Ebar = H.e_zero
        self.Exc = H.e_xc
        self.Etot = self.Ekin + self.e_coulomb + self.Ebar + self.Exc

    def get_td_energy(self):
        """Calculate the time-dependent total energy"""

        if not self.calculate_energy:
            self.Etot = 0.0
            return 0.0

        self.td_overlap.update(self.wfs)
        self.td_density.update()
        self.td_hamiltonian.update(self.td_density.get_density(),
                                   self.time)
        self.update_eigenvalues()

        return self.Etot

    def set_absorbing_boundary(self, absorbing_boundary):
        self.td_hamiltonian.set_absorbing_boundary(absorbing_boundary)

    # exp(ip.r) psi
    def absorption_kick(self, kick_strength):
        """Delta absorption kick for photoabsorption spectrum.

        Parameters:

        kick_strength: [float, float, float]
            Strength of the kick, e.g., [0.0, 0.0, 1e-3]

        """
        if self.rank == 0:
            self.text('Delta kick = ', kick_strength)

        self.kick_strength = np.array(kick_strength)

        abs_kick_hamiltonian = AbsorptionKickHamiltonian(
            self.wfs, self.spos_ac,
            np.array(kick_strength, float))
        abs_kick = AbsorptionKick(self.wfs, abs_kick_hamiltonian,
                                  self.td_overlap, self.solver,
                                  self.preconditioner, self.wfs.gd, self.timer)
        abs_kick.kick()

        # Kick the classical part, if it is present
        if self.hamiltonian.poisson.get_description() == 'FDTD+TDDFT':
            self.hamiltonian.poisson.set_kick(kick=self.kick_strength)



# Creates: planaraverages_BN.png

import numpy as np
import matplotlib.pyplot as plt
from ase.units import Bohr

data = np.load('electrostatic_data_BN_22.npz')
z = data['z'] * Bohr
V_model = data['V_model']
V_diff = data['V_X'] - data['V_0']
plt.plot(z, V_model.real, '-', label='$V(z)$')
plt.plot(z, V_diff.real[0], '-',
         label=(r'$[V^{V_\mathrm{C_B}^{+1}}_\mathrm{el}(z) -'
                r'V^{0}_\mathrm{el}(z) ]$'))

plt.plot(z, -V_diff.real[0] + V_model.real, '-',
         label='Difference')

constant = data['D_V_mean']
plt.axhline(constant, ls='dashed')
plt.axhline(0.0, ls='-', color='grey')
plt.xlabel(r'$z\enspace (\mathrm{\AA})$', fontsize=18)
plt.ylabel('Planar averages (eV)', fontsize=18)
plt.legend(loc='upper right')
plt.xlim((z[0], z[-1]))
plt.savefig('planaraverages_BN.png', bbox_inches='tight', dpi=300)



def create_tasks():
    from myqueue.task import task
    return [
        task('surface.agts.py'),
        task('work_function.py', deps='surface.agts.py')]


if __name__ == '__main__':
    from pathlib import Path
    source = Path('Al100.py').read_text()
    source = source.replace('k = ...', 'k = 6')
    source = source.replace('N = ...', 'N = 5')
    exec(source)



import numpy as np
import matplotlib.pyplot as plt

from gpaw import GPAW, setup_paths
from gpaw.xas import XAS

setup_paths.insert(0, '.')

h = 0.2

offset = 0.0
for L in np.arange(4, 14, 2) * 8 * h:
    calc = GPAW(f'h2o_hch_{L:.1f}.gpw')
    xas = XAS(calc)
    x, y = xas.get_spectra(fwhm=0.4)
    plt.plot(x, sum(y) + offset, label=f'{L:.1f}')
    offset += 0.005

plt.legend()
plt.xlim(-6, 4)
plt.ylim(-0.002, 0.03)
plt.savefig('h2o_xas_box.png')



from ase import Atom, Atoms

from gpaw import GPAW
from gpaw.analyse.overlap import Overlap
from gpaw.lrtddft.kssingle import KSSingles
from gpaw.test import equal

txt = '-'
txt = None
load = True
load = False
xc = 'LDA'

# run

R = 0.7  # approx. experimental bond length
a = 4.0
c = 5.0


class GPAW_with_classmethod_read(GPAW):
    @classmethod
    def read(cls, filename):
        gpw = cls()
        GPAW.read(gpw, filename)
        return gpw


def test_lrtddft_placzek_profeta_albrecht(in_tmp_dir):
    from ase.vibrations.albrecht import Albrecht
    from ase.vibrations.placzek import Placzek, Profeta
    from ase.vibrations.resonant_raman import ResonantRamanCalculator

    H2 = Atoms([Atom('H', (a / 2, a / 2, (c - R) / 2)),
                Atom('H', (a / 2, a / 2, (c + R) / 2))],
               cell=(a, a, c))

    name = exname = 'rraman'
    exkwargs = {'restrict': {'eps': 0.0, 'jend': 3}}

    if 1:
        calc = GPAW_with_classmethod_read(
            xc=xc, nbands=7,
            convergence={'bands': 3},
            spinpol=False,
            # eigensolver='rmm-diis',
            symmetry={'point_group': False},
            txt=txt)
        H2.calc = calc
        # H2.get_potential_energy()

        rr = ResonantRamanCalculator(
            H2, KSSingles, name=name, exname=exname,
            exkwargs=exkwargs,
            # XXX full does not work in parallel due to boxes
            # on different nodes
            # overlap=lambda x, y: Overlap(x).full(y)[0],
            overlap=lambda x, y: Overlap(x).pseudo(y)[0],
            txt=txt)
        rr.run()

    # check

    # Different Placzeck implementations should agree

    om = 5
    pz = Placzek(H2, KSSingles, name=name, exname=exname, txt=txt)
    pzi = pz.get_absolute_intensities(omega=om)[-1]

    pr = Profeta(H2, KSSingles, name=name, exname=exname,
                 approximation='Placzek', txt=txt)
    pri = pr.get_absolute_intensities(omega=om)[-1]
    equal(pzi, pri, 0.1)

    pr = Profeta(H2, KSSingles, name=name, exname=exname,
                 overlap=True,
                 approximation='Placzek', txt=txt)
    pri = pr.get_absolute_intensities(omega=om)[-1]
    equal(pzi, pri, 0.1)

    """Albrecht and Placzek are approximately equal"""

    al = Albrecht(H2, KSSingles, name=name, exname=exname,
                  overlap=True,
                  approximation='Albrecht', txt=txt)
    ali = al.get_absolute_intensities(omega=om)[-1]
    equal(pzi, ali, 1.5)

    """Albrecht A and P-P are approximately equal"""

    pr = Profeta(H2, KSSingles, name=name, exname=exname,
                 overlap=True,
                 approximation='P-P', txt=txt)
    pri = pr.get_absolute_intensities(omega=om)[-1]

    al = Albrecht(H2, KSSingles, name=name, exname=exname,
                  overlap=True,
                  approximation='Albrecht A', txt=txt)
    ali = al.get_absolute_intensities(omega=om)[-1]
    equal(pri, ali, 3)

    """Albrecht B+C and Profeta are approximately equal"""

    pr = Profeta(H2, KSSingles, name=name, exname=exname,
                 overlap=True,
                 approximation='Profeta', txt=txt)
    pri = pr.get_absolute_intensities(omega=om)[-1]

    al = Albrecht(H2, KSSingles, name=name, exname=exname,
                  overlap=True,
                  approximation='Albrecht BC', txt=txt)
    ali = al.get_absolute_intensities(omega=om)[-1]
    equal(pri, ali, 3)



#!/usr/bin/env python
from os import environ

label = 'oscillator'

octopus = environ['OCTOPUS_SCRIPT']
locals = {'label': label}
exec(open(octopus).read(), {}, locals)
exitcode = locals['exitcode']
if exitcode != 0:
    raise RuntimeError(('Octopus exited with exit code: %d.  ' +
                        'Check %s.log for more information.') %
                       (exitcode, label))



# Creates: Ag.png
from ase.build import bulk
from gpaw import GPAW

# Perform standard ground state calculation (with plane wave basis)
ag = bulk('Ag')
calc = GPAW(mode='pw',
            xc='LDA',
            kpts=(10, 10, 10),
            txt='Ag_LDA.txt')
ag.calc = calc
ag.get_potential_energy()
calc.write('Ag_LDA.gpw')

# Restart from ground state and fix potential:
calc = GPAW('Ag_LDA.gpw').fixed_density(
    nbands=16,
    basis='dzp',
    symmetry='off',
    convergence={'bands': 12},
    kpts={'path': 'WLGXWK', 'npoints': 100})

# Plot the band structure
band_structure = calc.band_structure()
band_structure.plot(filename='Ag.png', emax=20.0, show=True)



from gpaw import GPAW
from ase.build import molecule
from ase.units import mol, kJ, kcal, Pascal, m
from ase.data.vdw import vdw_radii
from ase.parallel import parprint
from gpaw.solvation import (
    SolvationGPAW,             # the solvation calculator
    EffectivePotentialCavity,  # cavity using an effective potential
    Power12Potential,          # a specific effective potential
    LinearDielectric,  # rule to construct permittivity func from the cavity
    GradientSurface,  # rule to calculate the surface area from the cavity
    SurfaceInteraction  # rule to calculate non-electrostatic interactions
)

# all parameters on the user side of the solvation API follow the ASE
# unit conventions (eV, Angstrom, ...)

# non-solvent related DFT parameters
h = 0.2
vac = 5.0

# solvent parameters for water from J. Chem. Phys. 141, 174108 (2014)
u0 = 0.180  # eV
epsinf = 78.36  # dimensionless
gamma = 18.4 * 1e-3 * Pascal * m  # convert from dyne / cm to eV / Angstrom**2
T = 298.15  # Kelvin
vdw_radii = vdw_radii.copy()
vdw_radii[1] = 1.09

# atomic_radii expected by gpaw.solvation.Power12Potential have to be a
# callable mapping an Atoms object to an iterable of floats representing
# the atomic radii of the atoms in the same order as in the Atoms object
# (in Angstrom)


def atomic_radii(atoms):
    return [vdw_radii[n] for n in atoms.numbers]


# create Atoms object for ethanol and add vacuum
atoms = molecule('CH3CH2OH')
atoms.center(vacuum=vac)

# perform gas phase calculation
atoms.calc = GPAW(xc='PBE', h=h, txt='gasphase.txt')
Egasphase = atoms.get_potential_energy()

# perform calculation with continuum solvent model from
# J. Chem. Phys. 141, 174108 (2014)
atoms.calc = SolvationGPAW(
    xc='PBE', h=h, txt='water.txt',
    cavity=EffectivePotentialCavity(
        effective_potential=Power12Potential(atomic_radii, u0),
        temperature=T,
        surface_calculator=GradientSurface()),
    dielectric=LinearDielectric(epsinf=epsinf),
    interactions=[SurfaceInteraction(surface_tension=gamma)])
Ewater = atoms.get_potential_energy()

# calculate solvation Gibbs energy in various units
DGSol_eV = Ewater - Egasphase
DGSol_kJ_per_mol = DGSol_eV / (kJ / mol)
DGSol_kcal_per_mol = DGSol_eV / (kcal / mol)

parprint('calculated Delta Gsol = %.0f meV = %.1f kJ / mol = %.1f kcal / mol' %
         (DGSol_eV * 1000., DGSol_kJ_per_mol, DGSol_kcal_per_mol))



from myqueue.task import task


def create_tasks():
    return [
        task('Na2TDDFT.py@2:1h'),
        task('part2.py', deps='Na2TDDFT.py'),
        task('ground_state.py@8:15s'),
        task('spectrum.py', deps='ground_state.py')]



# creates: intro/intro.ipynb
# creates: batteries/batteries1.ipynb
# creates: batteries/batteries2.ipynb
# creates: batteries/batteries3.ipynb
# creates: catalysis/n2_on_metal.ipynb, catalysis/neb.ipynb
# creates: catalysis/vibrations.ipynb, catalysis/convergence.ipynb
# creates: magnetism/magnetism1.ipynb, magnetism/magnetism2.ipynb
# creates: magnetism/magnetism3.ipynb,
# creates: machinelearning/machinelearning.ipynb
# creates: photovoltaics/pv1.ipynb, photovoltaics/pv2.ipynb
# creates: photovoltaics/pv3.ipynb
from pathlib import Path
from gpaw.utilities.nbrun import py2ipynb


for path in Path().glob('*/*.py'):
    if path.read_text().startswith('# %%\n'):
        print(path)
        py2ipynb(path)



scalapack = True

mklpath ='/global/apps/intel/2013.1/mkl'
omppath ='/global/apps/openmpi/1.6.5/intel/13.1'
lxcpath ='/home/pcje/global/apps/libxc-2.2.1-1'

compiler = 'icc'

libraries = ['xc', 'mpi', 'mkl_scalapack_lp64', 'mkl_lapack95_lp64', 'mkl_intel_lp64', 'mkl_sequential', 'mkl_mc', 'mkl_core', 'mkl_def', 'mkl_intel_thread', 'iomp5']
library_dirs += [f'{omppath}/lib']
library_dirs += [f'{mklpath}/lib/intel64']
library_dirs += [f'{lxcpath}/lib']
include_dirs += [f'{omppath}/include']
include_dirs += [f'{mklpath}/include']
include_dirs += [f'{lxcpath}/include']

extra_link_args += [f'{mklpath}/lib/intel64/libmkl_blacs_openmpi_lp64.a', f'{mklpath}/lib/intel64/libmkl_blas95_lp64.a']

extra_compile_args += ['-O3', '-std=c99', '-fPIC', '-Wall']

define_macros += [('GPAW_NO_UNDERSCORE_CBLACS', '1')]
define_macros += [('GPAW_NO_UNDERSCORE_CSCALAPACK', '1')]

mpicompiler = 'mpicc'
mpilinker = mpicompiler



# Creates: C_GW_k8_extrap.png
import pickle
import numpy as np
from ase.parallel import paropen
import matplotlib.pyplot as plt

plt.figure(figsize=(6.5, 4.5))

ecuts = np.array([100, 200, 300, 400])
color = ['ro-', 'bo-', 'go-', 'ko-', 'co-', 'mo-', 'yo-']
direct_gap = np.zeros(4)

k = 8
for i, ecut in enumerate([100, 200, 300, 400]):
    data = pickle.load(paropen(f'C-g0w0_k{k}_ecut{ecut}_results.pckl', 'rb'))
    direct_gap[i] = data['qp'][0, 0, 1] - data['qp'][0, 0, 0]
plt.plot(1 / (ecuts**(3. / 2.)), direct_gap, 'ko-',
         label=f'({k}x{k}x{k}) k-points')

extrap_gap, slope = np.linalg.solve([[1, 1. / 300.**(3. / 2)],
                                     [1, 1. / 400.**(3. / 2)]],
                                    [direct_gap[2], direct_gap[3]])
xs = np.linspace(0, 1 / 400.**(3. / 2), 1000)
plt.plot(xs, extrap_gap + slope * xs, 'k--')

plt.xticks([1. / 100**(3. / 2), 1. / 200**(3. / 2), 1. / 400**(3. / 2), 0],
           [100, 200, 400, r'$\infty$'])
plt.xlabel('Cutoff energy (eV)')
plt.ylabel('Direct band gap (eV)')
plt.title('non-selfconsistent G0W0@LDA')
plt.legend(loc='upper right')
plt.savefig('C_GW_k8_extrap.png')
plt.show()



import pytest
from ase import Atoms
from gpaw import GPAW
from gpaw.basis_data import get_basis_name, parse_basis_name

# This test looks for differently named setups and basis sets to ascertain
# that correct filenames are constructed.  Generally speaking, the things
# it looks for do not exist; we just verify the filenames.


@pytest.mark.ci
def test_setup_basis_spec():
    system = Atoms('Na')
    system.center(vacuum=3.0)

    def check(setups, basis, refname):
        calc = GPAW(setups=setups, basis=basis)
        system.calc = calc
        try:
            calc.initialize(system)
        except RuntimeError as err:
            msg = str(err)
            fname = msg.split('"')[1]
            assert fname == refname, fname
        else:
            assert False

    check({}, 'hello', 'Na.hello.basis')
    check('hello', {}, 'Na.hello.LDA')
    check('hello', 'dzp', 'Na.hello.dzp.basis')
    check('hello', 'sz(dzp)', 'Na.hello.dzp.basis')
    check('hello', 'world.dzp', 'Na.hello.world.dzp.basis')
    check('hello', 'sz(world.dzp)', 'Na.hello.world.dzp.basis')
    check('paw', 'world.dzp', 'Na.world.dzp.basis')
    check('paw', 'sz(world.dzp)', 'Na.world.dzp.basis')

    for basisname in 'sz dz szp dzp dz2p dzdp tzqp dzsp'.split():
        zetacount, polarizationcount = parse_basis_name(basisname)
        normalized_name = get_basis_name(zetacount, polarizationcount)
        zetacount2, polarizationcount2 = parse_basis_name(normalized_name)
        assert zetacount == zetacount2
        assert polarizationcount == polarizationcount2
        if (polarizationcount < 2 and
            len(basisname) < 4 and basisname.isalpha()):
            assert normalized_name == basisname



# Compare H adsorption energies and magnetic moments for a 12-ZGNR
# It needs 'ZGNR12.traj', 'ZGNR12_H.traj' and 'H2.traj'

from ase.io import read

reference = {'e_ads': -1.5,  # reference from rev
             'magmom_C': 0.5,
             'magmom_H': 0.5}
en = {}
magmom = {}
for name in ['ZGNR12', 'ZGNR12_H', 'H2']:
    atoms = read(name + '.traj')
    en[name] = atoms.get_potential_energy()

e_ads = en['ZGNR12_H'] - en['ZGNR12'] - en['H2']
assert abs(e_ads - reference['e_ads']) < 0.001

atoms = read('ZGNR12_H.traj')
magmom_C = atoms.get_magnetic_moments()[1]
magmom_H = atoms.get_magnetic_moments()[25]
assert abs(magmom_C - reference['magmom_C']) < 0.001
assert abs(magmom_H - reference['magmom_H']) < 0.001



from math import pi, cos, sin
from ase import Atoms
from ase.parallel import paropen
from gpaw import GPAW, setup_paths, FermiDirac
setup_paths.insert(0, '.')

a = 12.0  # use a large cell

d = 0.9575
t = pi / 180 * 104.51
atoms = Atoms('OH2',
              [(0, 0, 0),
               (d, 0, 0),
               (d * cos(t), d * sin(t), 0)],
              cell=(a, a, a))
atoms.center()

calc1 = GPAW(h=0.2,
             txt='h2o_gs.txt',
             xc='PBE')
atoms.calc = calc1
e1 = atoms.get_potential_energy() + calc1.get_reference_energy()

calc2 = GPAW(h=0.2,
             txt='h2o_exc.txt',
             xc='PBE',
             charge=-1,
             spinpol=True,
             occupations=FermiDirac(0.0, fixmagmom=True),
             setups={0: 'fch1s'})
atoms[0].magmom = 1
atoms.calc = calc2
e2 = atoms.get_potential_energy() + calc2.get_reference_energy()

with paropen('dks.result', 'w') as fd:
    print('Energy difference:', e2 - e1, file=fd)



import os
from ase.vibrations import Vibrations
from gpaw import GPAW
from gpaw.mpi import world

h2o = GPAW('h2o.gpw', txt=None).get_atoms()

# Test restart:
vib = Vibrations(h2o)
vib.run()
vib.summary(method='frederiksen')

# Remove a displacement file and redo it:
if world.rank == 0:
    os.remove('vib.1z-.pckl')
world.barrier()
vib = Vibrations(h2o)
vib.run()
vib.summary(method='frederiksen')



# Creates: tas2_eps.png
from ase import Atoms
from ase.lattice.hexagonal import Hexagonal
import matplotlib.pyplot as plt

from gpaw import GPAW, PW, FermiDirac
from gpaw.response.df import DielectricFunction
from gpaw.mpi import world
from gpaw.bztools import find_high_symmetry_monkhorst_pack

# 1) Ground-state.

a = 3.314
c = 12.1

cell = Hexagonal(symbol='Ta', latticeconstant={'a': a, 'c': c}).get_cell()
atoms = Atoms(symbols='TaS2TaS2', cell=cell, pbc=(1, 1, 1),
              scaled_positions=[(0, 0, 1 / 4),
                                (2 / 3, 1 / 3, 1 / 8),
                                (2 / 3, 1 / 3, 3 / 8),
                                (0, 0, 3 / 4),
                                (-2 / 3, -1 / 3, 5 / 8),
                                (-2 / 3, -1 / 3, 7 / 8)])

calc = GPAW(mode=PW(600),
            xc='PBE',
            occupations=FermiDirac(width=0.01),
            kpts={'density': 5})

atoms.calc = calc
atoms.get_potential_energy()
calc.write('TaS2-gs.gpw')

# 2) Unoccupied bands

kpts = find_high_symmetry_monkhorst_pack('TaS2-gs.gpw', density=5.0)

responseGS = GPAW('TaS2-gs.gpw').fixed_density(
    kpts=kpts,
    parallel={'band': 1},
    nbands=60,
    convergence={'bands': 50})

responseGS.get_potential_energy()
responseGS.write('TaS2-gsresponse.gpw', 'all')

# 3) Dielectric function

df = DielectricFunction('TaS2-gsresponse.gpw', eta=25e-3, domega0=0.01,
                        integrationmode='tetrahedron integration')

df1tetra_w, df2tetra_w = df.get_dielectric_function(direction='x')

df = DielectricFunction('TaS2-gsresponse.gpw', eta=25e-3,
                        domega0=0.01)
df1_w, df2_w = df.get_dielectric_function(direction='x')
omega_w = df.get_frequencies()

if world.rank == 0:
    plt.figure(figsize=(6, 6))
    plt.plot(omega_w, df2tetra_w.real, label='tetra Re')
    plt.plot(omega_w, df2tetra_w.imag, label='tetra Im')
    plt.plot(omega_w, df2_w.real, label='Re')
    plt.plot(omega_w, df2_w.imag, label='Im')
    plt.xlabel('Frequency (eV)')
    plt.ylabel('$\\varepsilon$')
    plt.xlim(0, 10)
    plt.ylim(-20, 20)
    plt.legend()
    plt.tight_layout()
    plt.savefig('tas2_eps.png', dpi=600)
#    plt.show()



import numpy as np
import pylab as plt

from gpaw import GPAW

# Read in the 5-layer slab:
calc = GPAW('slab-5.gpw', txt=None)
slab = calc.get_atoms()

# Get the height of the unit cell:
L = slab.get_cell()[2, 2]

# Get the effective potential on a 3D grid:
v = calc.get_effective_potential()

nx, ny, nz = v.shape
z = np.linspace(0, L, nz, endpoint=False)

efermi = calc.get_fermi_level()

# Calculate xy averaged potential:
vz = v.mean(axis=0).mean(axis=0)
print(f'Work function: {vz.max() - efermi:.2f} eV')

plt.plot(z, vz, label='xy averaged effective potential')
plt.plot([0, L], [efermi, efermi], label='Fermi level')
plt.ylabel('Potential / V')
plt.xlabel('z / Angstrom')
plt.legend(loc=0)
# plt.savefig('workfunction.png', format='png')
plt.show()



from ase.build import bulk

from gpaw import GPAW, FermiDirac
from gpaw.wavefunctions.pw import PW
from gpaw.response.g0w0 import G0W0

a = 3.567
atoms = bulk('C', 'diamond', a=a)

for k in [6, 8, 10, 12]:
    calc = GPAW(mode=PW(600),
                parallel={'domain': 1},
                kpts={'size': (k, k, k), 'gamma': True},
                xc='LDA',
                basis='dzp',
                occupations=FermiDirac(0.001),
                txt=f'C_groundstate_{k}.txt')

    atoms.calc = calc
    atoms.get_potential_energy()

    calc.diagonalize_full_hamiltonian()
    calc.write(f'C_groundstate_{k}.gpw', mode='all')

    for i, ecut in enumerate([100, 200, 300, 400]):
        gw = G0W0(calc=f'C_groundstate_{k}.gpw',
                  bands=(3, 5),
                  ecut=ecut,
                  kpts=[0],
                  filename=f'C-g0w0_k{k}_ecut{ecut}')

        result = gw.calculate()



import numpy as np
from ase.parallel import paropen
from gpaw import GPAW

atoms = GPAW('WS2_gs.gpw', txt=None).atoms

G, K, M = np.array([[0, 0, 0],
                    [1 / 3, 1 / 3, 0],
                    [0.5, 0, 0]])
bp = atoms.cell.bandpath([M, K, G, -K, -M], npoints=1000)

calc = GPAW('WS2_gs.gpw').fixed_density(
    kpts=bp.kpts, symmetry='off')
calc.write('WS2_bands.gpw')

x, X, labels = bp.get_linear_kpoint_axis()

with paropen('WS2_kpath.dat', 'w') as f:
    for k in x:
        print(k, file=f)


with paropen('WS2_highsym.dat', 'w') as f:
    for k in X:
        print(k, file=f)



import warnings
from math import sqrt, pi
import numpy as np

from ase.units import Ha
from gpaw.mpi import world
from gpaw.density import redistribute_array, redistribute_atomic_matrices
from gpaw.sphere.lebedev import weight_n
from gpaw.utilities import pack, pack_atomic_matrices, unpack_atomic_matrices
from gpaw.xc.gllb import safe_sqr
from gpaw.xc.gllb.contribution import Contribution

# XXX Work in process
debug = False


def d(*args):
    if debug:
        print(args)


class ResponsePotential:
    """Container for response potential"""
    def __init__(self, response, vt_sG, vt_sg, D_asp, Dresp_asp):
        self.response = response
        self.vt_sG = vt_sG
        self.vt_sg = vt_sg
        self.D_asp = D_asp
        self.Dresp_asp = Dresp_asp

    def redistribute(self, new_response):
        old_response = self.response
        new_wfs = new_response.wfs
        new_density = new_response.density
        self.vt_sG = redistribute_array(self.vt_sG,
                                        old_response.density.gd,
                                        new_density.gd,
                                        new_wfs.nspins,
                                        new_wfs.kptband_comm)
        if self.vt_sg is not None:
            self.vt_sg = redistribute_array(self.vt_sg,
                                            old_response.density.finegd,
                                            new_density.finegd,
                                            new_wfs.nspins,
                                            new_wfs.kptband_comm)

        def redist_asp(D_asp):
            return redistribute_atomic_matrices(D_asp,
                                                new_density.gd,
                                                new_wfs.nspins,
                                                new_density.setups,
                                                new_density.atom_partition,
                                                new_wfs.kptband_comm)

        self.D_asp = redist_asp(self.D_asp)
        self.Dresp_asp = redist_asp(self.Dresp_asp)
        self.response = new_response


class C_Response(Contribution):
    def __init__(self, weight, coefficients, damp=1e-10):
        Contribution.__init__(self, weight)
        d('In c_Response __init__', self)
        self.coefficients = coefficients
        self.vt_sg = None
        self.vt_sG = None
        self.D_asp = None
        self.Dresp_asp = None
        self.Ddist_asp = None
        self.Drespdist_asp = None
        self.damp = damp
        self.fix_potential = False

    def set_damp(self, damp):
        self.damp = damp

    def get_name(self):
        return 'RESPONSE'

    def get_desc(self):
        return self.coefficients.get_description()

    def initialize(self, density, hamiltonian, wfs):
        Contribution.initialize(self, density, hamiltonian, wfs)
        self.coefficients.initialize(wfs)
        if self.Dresp_asp is None:
            assert self.density.D_asp is None
        # XXX With the deprecated `fixdensity=True` option this
        # initialize() function is called both before AND after read()!
        # Thus, the second call of initialize() would discard the read
        # potential unless we check if it was already allocated.
        # Remove this if statement once `fixdensity=True` option has
        # been removed from the calculator.
        if self.vt_sg is None:
            self.vt_sG = self.gd.empty(self.nspins)
            self.vt_sg = self.finegd.empty(self.nspins)

    def initialize_1d(self, ae):
        Contribution.initialize_1d(self, ae)
        self.coefficients.initialize_1d(ae)

    def initialize_from_other_response(self, response):
        pot = ResponsePotential(response,
                                response.vt_sG,
                                response.vt_sg,
                                response.D_asp,
                                response.Dresp_asp)
        pot.redistribute(self)
        self.vt_sG = pot.vt_sG
        self.vt_sg = pot.vt_sg
        self.D_asp = pot.D_asp
        self.Dresp_asp = pot.Dresp_asp
        self.Ddist_asp = self.distribute_D_asp(pot.D_asp)
        self.Drespdist_asp = self.distribute_D_asp(pot.Dresp_asp)

    # Calcualte the GLLB potential and energy 1d
    def add_xc_potential_and_energy_1d(self, v_g):
        w_i = self.coefficients.get_coefficients_1d()
        u2_j = safe_sqr(self.ae.u_j)
        v_g += self.weight * np.dot(w_i, u2_j) / (
            np.dot(self.ae.f_j, u2_j) + self.damp)
        return 0.0  # Response part does not contribute to energy

    def update_potentials(self):
        d('In update response potential')
        if self.fix_potential:
            # Skip the evaluation of the potential so that
            # the existing potential is used. This is relevant for
            # band structure calculation so that the potential
            # does not get updated with the other k-point sampling.
            return

        if self.wfs.kpt_u[0].eps_n is None:
            # This skips the evaluation of the potential so that
            # the existing one is used.
            # This happens typically after reading when occupations
            # haven't been calculated yet and the potential read earlier
            # is used.
            return

        w_kn = self.coefficients.get_coefficients(self.wfs.kpt_u)
        f_kn = [kpt.f_n for kpt in self.wfs.kpt_u]
        if w_kn is not None:
            self.vt_sG[:] = 0.0
            nt_sG = self.gd.zeros(self.nspins)

            for kpt, w_n in zip(self.wfs.kpt_u, w_kn):
                self.wfs.add_to_density_from_k_point_with_occupation(
                    self.vt_sG, kpt, w_n)
                self.wfs.add_to_density_from_k_point(nt_sG, kpt)

            self.wfs.kptband_comm.sum(nt_sG)
            self.wfs.kptband_comm.sum(self.vt_sG)

            if self.wfs.kd.symmetry:
                for nt_G, vt_G in zip(nt_sG, self.vt_sG):
                    self.wfs.kd.symmetry.symmetrize(nt_G, self.gd)
                    self.wfs.kd.symmetry.symmetrize(vt_G, self.gd)

            d('response update D_asp', world.rank, self.Dresp_asp.keys(),
              self.D_asp.keys())
            self.wfs.calculate_atomic_density_matrices_with_occupation(
                self.Dresp_asp, w_kn)
            self.Drespdist_asp = self.distribute_D_asp(self.Dresp_asp)
            d('response update Drespdist_asp', world.rank,
              self.Dresp_asp.keys(), self.D_asp.keys())
            self.wfs.calculate_atomic_density_matrices_with_occupation(
                self.D_asp, f_kn)
            self.Ddist_asp = self.distribute_D_asp(self.D_asp)

            self.vt_sG /= nt_sG + self.damp

        self.density.distribute_and_interpolate(self.vt_sG, self.vt_sg)

    def calculate(self, e_g, n_sg, v_sg):
        self.update_potentials()
        v_sg += self.weight * self.vt_sg

    def distribute_D_asp(self, D_asp):
        return self.density.atomdist.to_work(D_asp)

    def calculate_energy_and_derivatives(self, setup, D_sp, H_sp, a,
                                         addcoredensity=True):
        # Get the XC-correction instance
        c = setup.xc_correction
        ncresp_g = setup.extra_xc_data['core_response'] / self.nspins
        if not addcoredensity:
            ncresp_g[:] = 0.0

        for D_p, dEdD_p, Dresp_p in zip(self.Ddist_asp[a], H_sp,
                                        self.Drespdist_asp[a]):
            D_Lq = np.dot(c.B_pqL.T, D_p)
            n_Lg = np.dot(D_Lq, c.n_qg)  # Construct density
            if addcoredensity:
                n_Lg[0] += c.nc_g * sqrt(4 * pi) / self.nspins
            nt_Lg = np.dot(
                D_Lq, c.nt_qg
            )  # Construct smooth density (_without smooth core density_)

            Dresp_Lq = np.dot(c.B_pqL.T, Dresp_p)
            nresp_Lg = np.dot(Dresp_Lq, c.n_qg)  # Construct 'response density'
            nrespt_Lg = np.dot(
                Dresp_Lq, c.nt_qg
            )  # Construct smooth 'response density' (w/o smooth core)

            for w, Y_L in zip(weight_n, c.Y_nL):
                nt_g = np.dot(Y_L, nt_Lg)
                nrespt_g = np.dot(Y_L, nrespt_Lg)
                x_g = nrespt_g / (nt_g + self.damp)
                dEdD_p -= self.weight * w * np.dot(
                    np.dot(c.B_pqL, Y_L), np.dot(c.nt_qg, x_g * c.rgd.dv_g))

                n_g = np.dot(Y_L, n_Lg)
                nresp_g = np.dot(Y_L, nresp_Lg)
                x_g = (nresp_g + ncresp_g) / (n_g + self.damp)

                dEdD_p += self.weight * w * np.dot(
                    np.dot(c.B_pqL, Y_L), np.dot(c.n_qg, x_g * c.rgd.dv_g))

        return 0.0

    def integrate_sphere(self, a, Dresp_sp, D_sp, Dwf_p, spin=0):
        c = self.wfs.setups[a].xc_correction
        Dresp_p, D_p = Dresp_sp[spin], D_sp[spin]
        D_Lq = np.dot(c.B_pqL.T, D_p)
        n_Lg = np.dot(D_Lq, c.n_qg)  # Construct density
        n_Lg[0] += c.nc_g * sqrt(4 * pi) / len(D_sp)
        nt_Lg = np.dot(D_Lq, c.nt_qg
                       )  # Construct smooth density (without smooth core)
        Dresp_Lq = np.dot(c.B_pqL.T, Dresp_p)  # Construct response
        nresp_Lg = np.dot(Dresp_Lq, c.n_qg)  # Construct 'response density'
        nrespt_Lg = np.dot(
            Dresp_Lq, c.nt_qg
        )  # Construct smooth 'response density' (w/o smooth core)
        Dwf_Lq = np.dot(c.B_pqL.T, Dwf_p)  # Construct lumo wf
        nwf_Lg = np.dot(Dwf_Lq, c.n_qg)
        nwft_Lg = np.dot(Dwf_Lq, c.nt_qg)
        E = 0.0
        for w, Y_L in zip(weight_n, c.Y_nL):
            v = np.dot(Y_L, nwft_Lg) * np.dot(Y_L, nrespt_Lg) / (
                np.dot(Y_L, nt_Lg) + self.damp)
            E -= self.weight * w * np.dot(v, c.rgd.dv_g)
            v = np.dot(Y_L, nwf_Lg) * np.dot(Y_L, nresp_Lg) / (
                np.dot(Y_L, n_Lg) + self.damp)
            E += self.weight * w * np.dot(v, c.rgd.dv_g)
        return E

    def add_smooth_xc_potential_and_energy_1d(self, vt_g):
        w_ln = self.coefficients.get_coefficients_1d(smooth=True)
        v_g = np.zeros(self.ae.N)
        n_g = np.zeros(self.ae.N)
        for w_n, f_n, u_n in zip(w_ln, self.ae.f_ln,
                                 self.ae.s_ln):  # For each angular momentum
            u2_n = safe_sqr(u_n)
            v_g += np.dot(w_n, u2_n)
            n_g += np.dot(f_n, u2_n)

        vt_g += self.weight * v_g / (n_g + self.damp)
        return 0.0  # Response part does not contribute to energy

    def calculate_delta_xc(self, homolumo=None):
        warnings.warn(
            'The function calculate_delta_xc() is deprecated. '
            'Please use calculate_discontinuity_potential() instead. '
            'See documentation on calculating band gap with GLLBSC.',
            DeprecationWarning)

        if homolumo is None:
            # Calculate band gap

            # This always happens, so we don't warn.
            # We should perhaps print it as an ordinary message,
            # but we do not have a file here to which to print.
            # import warnings
            # warnings.warn('Calculating KS-gap directly from the k-points, '
            #               'can be inaccurate.')
            homolumo = self.wfs.get_homo_lumo()
        homo, lumo = homolumo
        dxc_pot = self.calculate_discontinuity_potential(homo * Ha, lumo * Ha)
        self.Dxc_vt_sG = dxc_pot.vt_sG
        self.Dxc_D_asp = dxc_pot.D_asp
        self.Dxc_Dresp_asp = dxc_pot.Dresp_asp

    def calculate_discontinuity_potential(self, homo, lumo):
        r"""Calculate the derivative discontinuity potential.

        This function calculates $`\Delta_{xc}(r)`$ as given by
        Eq. (24) of https://doi.org/10.1103/PhysRevB.82.115106

        Parameters:

        homo:
            homo energy (or energies in spin-polarized case) in eV
        lumo:
            lumo energy (or energies in spin-polarized case) in eV

        Returns:

        dxc_pot: Discontinuity potential
        """
        homolumo = np.asarray((homo, lumo)) / Ha

        dxc_Dresp_asp = self.empty_atomic_matrix()
        dxc_D_asp = self.empty_atomic_matrix()
        for a in self.density.D_asp:
            ni = self.wfs.setups[a].ni
            dxc_Dresp_asp[a] = np.zeros((self.nspins, ni * (ni + 1) // 2))
            dxc_D_asp[a] = np.zeros((self.nspins, ni * (ni + 1) // 2))

        # Calculate new response potential with LUMO reference
        w_kn = self.coefficients.get_coefficients_for_lumo_perturbation(
            self.wfs.kpt_u,
            homolumo=homolumo)
        f_kn = [kpt.f_n for kpt in self.wfs.kpt_u]

        dxc_vt_sG = self.gd.zeros(self.nspins)
        nt_sG = self.gd.zeros(self.nspins)
        for kpt, w_n in zip(self.wfs.kpt_u, w_kn):
            self.wfs.add_to_density_from_k_point_with_occupation(dxc_vt_sG,
                                                                 kpt, w_n)
            self.wfs.add_to_density_from_k_point(nt_sG, kpt)

        self.wfs.kptband_comm.sum(nt_sG)
        self.wfs.kptband_comm.sum(dxc_vt_sG)

        if self.wfs.kd.symmetry:
            for nt_G, dxc_vt_G in zip(nt_sG, dxc_vt_sG):
                self.wfs.kd.symmetry.symmetrize(nt_G, self.gd)
                self.wfs.kd.symmetry.symmetrize(dxc_vt_G, self.gd)

        dxc_vt_sG /= nt_sG + self.damp

        self.wfs.calculate_atomic_density_matrices_with_occupation(
            dxc_Dresp_asp, w_kn)
        self.wfs.calculate_atomic_density_matrices_with_occupation(
            dxc_D_asp, f_kn)
        dxc_pot = ResponsePotential(self, dxc_vt_sG, None,
                                    dxc_D_asp, dxc_Dresp_asp)
        return dxc_pot

    def calculate_delta_xc_perturbation(self):
        warnings.warn(
            'The function calculate_delta_xc_perturbation() is deprecated. '
            'Please use calculate_discontinuity() instead. '
            'See documentation on calculating band gap with GLLBSC.',
            DeprecationWarning)
        dxc_pot = ResponsePotential(self, self.Dxc_vt_sG, None, self.Dxc_D_asp,
                                    self.Dxc_Dresp_asp)
        if self.nspins != 1:
            ret = []
            for spin in range(self.nspins):
                ret.append(self.calculate_discontinuity(dxc_pot, spin=spin))
            return ret
        return self.calculate_discontinuity(dxc_pot)

    def calculate_delta_xc_perturbation_spin(self, s=0):
        warnings.warn(
            'The function calculate_delta_xc_perturbation_spin() is '
            'deprecated. Please use calculate_discontinuity_spin() instead. '
            'See documentation on calculating band gap with GLLBSC.',
            DeprecationWarning)
        dxc_pot = ResponsePotential(self, self.Dxc_vt_sG, None, self.Dxc_D_asp,
                                    self.Dxc_Dresp_asp)
        return self.calculate_discontinuity(dxc_pot, spin=s)

    def calculate_discontinuity(self,
                                dxc_pot: ResponsePotential,
                                spin: int = None):
        r"""Calculate the derivative discontinuity.

        This function evaluates the perturbation theory expression
        for the derivative discontinuity value as given by
        Eq. (25) of https://doi.org/10.1103/PhysRevB.82.115106

        Parameters:

        dxc_pot:
            Discontinuity potential.
        spin:
            Spin value.

        Returns:

        KSgap:
            Kohn-Sham gap in eV
        dxc:
            Derivative discontinuity in eV
        """
        if spin is None:
            if self.nspins != 1:
                raise ValueError('Specify spin for the discontinuity.')
            spin = 0

        # Redistribute discontinuity potential
        if dxc_pot.response is not self:
            dxc_pot.redistribute(self)

        homo, lumo = self.wfs.get_homo_lumo(spin)
        KSgap = lumo - homo

        # Find the lumo-orbital of this spin
        eps_n = self.wfs.kpt_qs[0][spin].eps_n
        assert self.wfs.bd.comm.size == 1
        lumo_n = (eps_n < self.wfs.fermi_level).sum()

        # Calculate the expectation value for all k points, and keep
        # the smallest energy value
        nt_G = self.gd.empty()
        min_energy = np.inf
        for u, kpt in enumerate(self.wfs.kpt_u):
            if kpt.s == spin:
                nt_G[:] = 0.0
                self.wfs.add_orbital_density(nt_G, kpt, lumo_n)
                E = 0.0
                for a in dxc_pot.D_asp:
                    D_sp = dxc_pot.D_asp[a]
                    Dresp_sp = dxc_pot.Dresp_asp[a]
                    P_ni = kpt.P_ani[a]
                    Dwf_p = pack(np.outer(P_ni[lumo_n].T.conj(),
                                          P_ni[lumo_n]).real)
                    E += self.integrate_sphere(a, Dresp_sp, D_sp, Dwf_p,
                                               spin=spin)
                E = self.gd.comm.sum(E)
                E += self.gd.integrate(nt_G * dxc_pot.vt_sG[spin])
                E += kpt.eps_n[lumo_n] - lumo
                min_energy = min(min_energy, E)

        # Take the smallest value over all distributed k points
        dxc = self.wfs.kd.comm.min(min_energy)
        return KSgap * Ha, dxc * Ha

    def calculate_discontinuity_from_average(self,
                                             dxc_pot: ResponsePotential,
                                             spin: int,
                                             testing: bool = False):
        msg = ('This function estimates discontinuity by calculating '
               'the average of the discontinuity potential. '
               'Use only for testing and debugging.')
        if not testing:
            raise RuntimeError(msg)
        else:
            warnings.warn(msg)
        assert self.wfs.world.size == 1

        # Calculate average of lumo reference response potential
        dxc = np.average(dxc_pot.vt_sG[spin])
        return dxc * Ha

    def initialize_from_atomic_orbitals(self, basis_functions):
        # Initialize 'response-density' and density-matrices
        # print('Initializing from atomic orbitals')
        self.Dresp_asp = self.empty_atomic_matrix()
        setups = self.wfs.setups

        for a in self.density.D_asp.keys():
            ni = setups[a].ni
            self.Dresp_asp[a] = np.zeros((self.nspins, ni * (ni + 1) // 2))

        self.D_asp = self.empty_atomic_matrix()
        f_asi = {}
        w_asi = {}

        for a in basis_functions.atom_indices:
            if setups[a].type == 'ghost':
                w_j = [0.0]
            else:
                w_j = setups[a].extra_xc_data['w_j']

            # Basis function coefficients based of response weights
            w_si = setups[a].calculate_initial_occupation_numbers(
                0, False,
                charge=0,
                nspins=self.nspins,
                f_j=w_j)
            # Basis function coefficients based on density
            f_si = setups[a].calculate_initial_occupation_numbers(
                0, False,
                charge=0,
                nspins=self.nspins)

            if a in basis_functions.my_atom_indices:
                self.Dresp_asp[a] = setups[a].initialize_density_matrix(w_si)
                self.D_asp[a] = setups[a].initialize_density_matrix(f_si)

            f_asi[a] = f_si
            w_asi[a] = w_si

        self.Drespdist_asp = self.distribute_D_asp(self.Dresp_asp)
        self.Ddist_asp = self.distribute_D_asp(self.D_asp)
        nt_sG = self.gd.zeros(self.nspins)
        basis_functions.add_to_density(nt_sG, f_asi)
        self.vt_sG = self.gd.zeros(self.nspins)
        basis_functions.add_to_density(self.vt_sG, w_asi)
        # Update vt_sG to correspond atomic response potential. This will be
        # used until occupations and eigenvalues are available.
        self.vt_sG /= nt_sG + self.damp
        self.vt_sg = self.finegd.zeros(self.nspins)
        self.density.distribute_and_interpolate(self.vt_sG, self.vt_sg)

    def get_extra_setup_data(self, extra_data):
        ae = self.ae
        njcore = ae.njcore
        w_ln = self.coefficients.get_coefficients_1d(smooth=True)
        w_j = []
        for w_n in w_ln:
            for w in w_n:
                w_j.append(w)
        extra_data['w_j'] = w_j

        w_j = self.coefficients.get_coefficients_1d()
        x_g = np.dot(w_j[:njcore], safe_sqr(ae.u_j[:njcore]))
        x_g[1:] /= ae.r[1:] ** 2 * 4 * np.pi
        x_g[0] = x_g[1]
        extra_data['core_response'] = x_g

        # For debugging purposes
        w_j = self.coefficients.get_coefficients_1d()
        u2_j = safe_sqr(self.ae.u_j)
        v_g = self.weight * np.dot(w_j, u2_j) / (
            np.dot(self.ae.f_j, u2_j) + self.damp)
        v_g[0] = v_g[1]
        extra_data['all_electron_response'] = v_g

        # Calculate Hardness of spherical atom, for debugging purposes
        l = [np.where(f < 1e-3, e, 1000)
             for f, e in zip(self.ae.f_j, self.ae.e_j)]
        h = [np.where(f > 1e-3, e, -1000)
             for f, e in zip(self.ae.f_j, self.ae.e_j)]
        lumo_e = min(l)
        homo_e = max(h)
        if lumo_e < 999:  # If there is unoccpied orbital
            w_j = self.coefficients.get_coefficients_1d_for_lumo_perturbation()
            v_g = self.weight * np.dot(w_j, u2_j) / (
                np.dot(self.ae.f_j, u2_j) + self.damp)
            e2 = [e + np.dot(u2 * v_g, self.ae.dr)
                  for u2, e in zip(u2_j, self.ae.e_j)]
            lumo_2 = min([np.where(f < 1e-3, e, 1000)
                          for f, e in zip(self.ae.f_j, e2)])
            # print('New lumo eigenvalue:', lumo_2 * 27.2107)
            self.hardness = lumo_2 - homo_e
            # print('Hardness predicted: %10.3f eV' %
            #       (self.hardness * 27.2107))

    def write(self, writer):
        """Writes response specific data."""
        wfs = self.wfs
        kpt_comm = wfs.kd.comm
        gd = wfs.gd

        # Write the pseudodensity on the coarse grid:
        shape = (wfs.nspins,) + tuple(gd.get_size_of_global_array())
        writer.add_array('gllb_pseudo_response_potential', shape)
        if kpt_comm.rank == 0:
            for vt_G in self.vt_sG:
                writer.fill(gd.collect(vt_G) * Ha)

        def pack(X0_asp):
            X_asp = wfs.setups.empty_atomic_matrix(
                wfs.nspins, wfs.atom_partition)
            # XXX some of the provided X0_asp contain strangely duplicated
            # elements.  Take only the minimal set:
            for a in X_asp:
                X_asp[a][:] = X0_asp[a]
            return pack_atomic_matrices(X_asp)

        writer.write(gllb_atomic_density_matrices=pack(self.D_asp))
        writer.write(gllb_atomic_response_matrices=pack(self.Dresp_asp))

    def empty_atomic_matrix(self):
        assert self.wfs.atom_partition is self.density.atom_partition
        return self.wfs.setups.empty_atomic_matrix(self.wfs.nspins,
                                                   self.wfs.atom_partition)

    def read(self, reader):
        r = reader.hamiltonian.xc
        wfs = self.wfs

        d('Reading vt_sG')
        self.gd.distribute(r.gllb_pseudo_response_potential / reader.ha,
                           self.vt_sG)
        self.density.distribute_and_interpolate(self.vt_sG, self.vt_sg)

        def unpack(D_sP):
            return unpack_atomic_matrices(D_sP, wfs.setups)

        # Read atomic density matrices and non-local part of hamiltonian:
        D_asp = unpack(r.gllb_atomic_density_matrices)
        Dresp_asp = unpack(r.gllb_atomic_response_matrices)

        # All density matrices are loaded to all cores
        # First distribute them to match density.D_asp
        self.D_asp = self.empty_atomic_matrix()
        self.Dresp_asp = self.empty_atomic_matrix()
        for a in self.density.D_asp:
            self.D_asp[a][:] = D_asp[a]
            self.Dresp_asp[a][:] = Dresp_asp[a]
        assert len(self.Dresp_asp) == len(self.density.D_asp)

        # Further distributes the density matricies for xc-corrections
        self.Ddist_asp = self.distribute_D_asp(self.D_asp)
        self.Drespdist_asp = self.distribute_D_asp(self.Dresp_asp)

    def heeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeelp(self, olddens):
        # XXX This function should be removed once the deprecated
        # `fixdensity=True` option is removed.
        from gpaw.density import redistribute_array
        self.vt_sg = redistribute_array(self.vt_sg,
                                        olddens.finegd, self.finegd,
                                        self.wfs.nspins, self.wfs.kptband_comm)
        self.vt_sG = redistribute_array(self.vt_sG,
                                        olddens.gd, self.gd,
                                        self.wfs.nspins, self.wfs.kptband_comm)



import os

import numpy as np
import pytest

from gpaw import GPAW, PW
from gpaw.mpi import world
from gpaw.test import equal, findpeak
from gpaw.response.df import DielectricFunction
from ase.build import bulk
from ase.units import Bohr, Hartree


@pytest.mark.slow
def test_chi0_intraband(in_tmp_dir):
    """Comparing the plasmon peaks found in bulk sodium for two different
    atomic structures. Testing for idential plasmon peaks. Not using
    physical sodium cell."""
    a1 = bulk('Na')
    a2 = bulk('Na')
    a2.set_initial_magnetic_moments([[0.1, ]])
    a1.calc = GPAW(mode=PW(300),
                   kpts={'size': (8, 8, 8), 'gamma': True},
                   parallel={'band': 1},
                   txt='na_spinpaired.txt')
    a2.calc = GPAW(mode=PW(300),
                   kpts={'size': (8, 8, 8), 'gamma': True},
                   parallel={'band': 1},
                   txt='na_spinpol.txt')
    a1.get_potential_energy()
    a2.get_potential_energy()

    # Use twice as many bands for expanded structure
    a1.calc.diagonalize_full_hamiltonian(nbands=20)
    a2.calc.diagonalize_full_hamiltonian(nbands=20)

    a1.calc.write('intraband_spinpaired.gpw', 'all')
    a2.calc.write('intraband_spinpolarized.gpw', 'all')

    # Calculate the dielectric functions
    if world.rank == 0:
        try:
            os.remove('intraband_spinpaired+0+0+0.pckl')
        except OSError:
            pass

    df1 = DielectricFunction('intraband_spinpaired.gpw',
                             domega0=0.03,
                             ecut=10,
                             rate=0.1,
                             integrationmode='tetrahedron integration',
                             name='intraband_spinpaired',
                             txt='intraband_spinpaired_df.txt')

    df1NLFCx, df1LFCx = df1.get_dielectric_function(direction='x')
    df1NLFCy, df1LFCy = df1.get_dielectric_function(direction='y')
    df1NLFCz, df1LFCz = df1.get_dielectric_function(direction='z')
    wp1_vv = df1.chi0.plasmafreq_vv**0.5
    wp1 = wp1_vv[0, 0]
    if world.rank == 0:
        try:
            os.remove('intraband_spinpolarized+0+0+0.pckl')
        except OSError:
            pass

    df2 = DielectricFunction('intraband_spinpolarized.gpw',
                             domega0=0.03,
                             ecut=10,
                             rate=0.1,
                             integrationmode='tetrahedron integration',
                             name='intraband_spinpolarized',
                             txt='intraband_spinpolarized_df.txt')

    df2NLFCx, df2LFCx = df2.get_dielectric_function(direction='x')
    df2NLFCy, df2LFCy = df2.get_dielectric_function(direction='y')
    df2NLFCz, df2LFCz = df2.get_dielectric_function(direction='z')
    wp2_vv = df2.chi0.plasmafreq_vv**0.5
    wp2 = wp2_vv[0, 0]

    # Compare plasmon frequencies and intensities
    w_w = df1.chi0.omega_w

    # Analytical Drude result
    n = 1 / (a1.get_volume() * Bohr**-3)

    wp = np.sqrt(4 * np.pi * n)

    # From http://pubs.acs.org/doi/abs/10.1021/jp810808h
    wpref = 5.71 / Hartree

    equal(wp1, wp2, 1e-2)
    equal(wp1, wp, 0.5)  # Use larger margin when comparing to Drude
    equal(wp1, wpref, 0.1)

    w1, I1 = findpeak(w_w, -(1. / df1LFCx).imag)
    w2, I2 = findpeak(w_w, -(1. / df2LFCx).imag)
    equal(w1, w2, 1e-2)
    equal(I1, I2, 1e-1)

    w1, I1 = findpeak(w_w, -(1. / df1LFCy).imag)
    w2, I2 = findpeak(w_w, -(1. / df2LFCy).imag)
    equal(w1, w2, 1e-2)
    equal(I1, I2, 1e-1)

    w1, I1 = findpeak(w_w, -(1. / df1LFCz).imag)
    w2, I2 = findpeak(w_w, -(1. / df2LFCz).imag)
    equal(w1, w2, 1e-2)
    equal(I1, I2, 1e-1)



from ase import Atoms
from ase.io import write
from gpaw import GPAW

# Oxygen atom:
atom = Atoms('O', cell=[6, 6, 6], pbc=False)
atom.center()

calc = GPAW(h=0.2,
            hund=True,  # assigns the atom its correct magnetic moment
            txt='O.txt')

atom.calc = calc
atom.get_potential_energy()

# Write wave functions to gpw file:
calc.write('O.gpw', mode='all')

# Generate cube-files of the orbitals:
for spin in [0, 1]:
    for n in range(calc.get_number_of_bands()):
        wf = calc.get_pseudo_wave_function(band=n, spin=spin)
        write('O.%d.%d.cube' % (spin, n), atom, data=wf)



import pytest
import numpy as np
from gpaw import GPAW, PW
from ase import Atoms
from gpaw.nlopt.shg import get_shg
from gpaw.nlopt.matrixel import make_nlodata
from gpaw.mpi import world


@pytest.mark.skipif(world.size > 4, reason='System too small')
def test_shg(in_tmp_dir):
    # Check for Hydrogen atom
    atoms = Atoms('H', cell=(3 * np.eye(3)), pbc=True)

    # Do a GS and save it
    calc = GPAW(
        mode=PW(600), symmetry={'point_group': False},
        kpts={'size': (2, 2, 2)}, nbands=5, txt=None)
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('gs.gpw', 'all')

    # Get the mml
    make_nlodata()

    # Do a SHG
    get_shg(freqs=np.linspace(0, 5, 100))
    get_shg(freqs=np.linspace(0, 5, 100), gauge='vg', out_name='shg2.npy')

    # Check it
    if world.rank == 0:
        shg = np.load('shg.npy')
        shg2 = np.load('shg2.npy')
        # Check for nan's
        assert not np.isnan(shg).any()
        assert not np.isnan(shg2).any()
        # Check the two gauges
        assert np.all(np.abs(shg2[1] - shg[1]) < 1e-3)
        # It should be zero (small) since H2 is centro-symm.
        assert np.all(np.abs(shg[1]) < 1e-8)



# 19 Nov 2009: problems with scalapack
#scalapack = True

extra_compile_args = ['-fast', '-std=c99', '-fPIC']

compiler = 'icc'

libraries = ['mkl_core', 'mkl_sequential', 'mkl_gf_lp64', 'iomp5']
#libraries = ['mkl_core', 'mkl_intel_thread', 'mkl_gf_lp64', 'mkl_blacs_intelmpi_lp64', 'mkl_scalapack_lp64', 'iomp5']

mkl_lib_path = '/software/intel/mkl/10.2.1.017/lib/em64t/'

library_dirs = [mkl_lib_path]

extra_link_args =['-Wl,-rpath='+mkl_lib_path]

define_macros += [('GPAW_NO_UNDERSCORE_CBLACS', '1')]
define_macros += [('GPAW_NO_UNDERSCORE_CSCALAPACK', '1')]



from ase import Atoms
from gpaw import GPAW

a = 5.475
b = a / 2
si = Atoms(symbols='Si4',
           positions=[(0, 0, 0),
                      (0, b, b),
                      (b, 0, b),
                      (b, b, 0)],
           cell=(a, a, a),
           pbc=True)
si += si
si.positions[4:] += a / 4

calc = GPAW(nbands=16,
            h=0.25,
            txt='si.txt')

si.calc = calc
si.get_potential_energy()
calc.write('si.gpw', mode='all')



# Creates: lcaodos.png
import matplotlib.pyplot as plt
import numpy as np
from ase.io import read
from ase.units import Hartree

from gpaw import GPAW
from gpaw.utilities.dos import RestartLCAODOS, fold


name = 'HfS2'
calc = GPAW(name + '.gpw', txt=None)
atoms = read(name + '.gpw')
ef = calc.get_fermi_level()

dos = RestartLCAODOS(calc)
energies, weights = dos.get_subspace_pdos(range(51))
e, w = fold(energies * Hartree, weights, 2000, 0.1)

e, m_s_pdos = dos.get_subspace_pdos([0, 1])
e, m_s_pdos = fold(e * Hartree, m_s_pdos, 2000, 0.1)
e, m_p_pdos = dos.get_subspace_pdos([2, 3, 4])
e, m_p_pdos = fold(e * Hartree, m_p_pdos, 2000, 0.1)
e, m_d_pdos = dos.get_subspace_pdos([5, 6, 7, 8, 9])
e, m_d_pdos = fold(e * Hartree, m_d_pdos, 2000, 0.1)

e, x_s_pdos = dos.get_subspace_pdos([25])
e, x_s_pdos = fold(e * Hartree, x_s_pdos, 2000, 0.1)
e, x_p_pdos = dos.get_subspace_pdos([26, 27, 28])
e, x_p_pdos = fold(e * Hartree, x_p_pdos, 2000, 0.1)

w_max = []
for i in range(len(e)):
    if (-4.5 <= e[i] - ef <= 4.5):
        w_max.append(w[i])

w_max = np.asarray(w_max)

# Plot things:
plt.plot(e - ef, w, label='Total', c='k', lw=2, alpha=0.7)
plt.plot(e - ef, x_s_pdos, label='X-s', c='g', lw=2, alpha=0.7)
plt.plot(e - ef, x_p_pdos, label='X-p', c='b', lw=2, alpha=0.7)
plt.plot(e - ef, m_s_pdos, label='M-s', c='y', lw=2, alpha=0.7)
plt.plot(e - ef, m_p_pdos, label='M-p', c='c', lw=2, alpha=0.7)
plt.plot(e - ef, m_d_pdos, label='M-d', c='r', lw=2, alpha=0.7)

plt.axis(ymin=0., ymax=np.max(w_max), xmin=-4.5, xmax=4.5, )
plt.xlabel(r'$\epsilon - \epsilon_F \ \rm{(eV)}$')
plt.ylabel('DOS')
plt.legend(loc=1)
plt.savefig('lcaodos.png')
plt.show()



# creates: organometal.db
from ase.db import connect
import random

in_db = connect('organometal.master.db')
out_db = connect('organometal.db')

preserved_names = ['MAPbI3', 'FASnBr2Cl', 'FASnBr3', 'FASnCl3']
rows = [row for row in in_db.select(project='organometal')]
names = set([row.name for row in rows])
for name in names:
    n = random.randint(1, 4)
    sub_rows = [x for x in rows if x.name == name]
    if name not in preserved_names:
        sub_rows = list(random.sample(sub_rows, n))
    for row in sub_rows:
        out_db.write(row, **row.key_value_pairs)

rows = [row for row in in_db.select(subproject='references')]
for row in rows:
    print(row)
    print(row.key_value_pairs)
    out_db.write(row, **row.key_value_pairs)
print()



from gpaw.response.g0w0 import G0W0
from ase.build import mx2
from gpaw import GPAW, PW, FermiDirac

structure = mx2(formula='MoS2', kind='2H', a=3.184, thickness=3.127,
                size=(1, 1, 1), vacuum=3.5)
structure.pbc = (1, 1, 1)

Ecut = 400

calc = GPAW(mode=PW(Ecut),
            parallel={'domain': 1},
            xc='PBE',
            basis='dzp',
            kpts={'size': (9, 9, 1), 'gamma': True},
            occupations=FermiDirac(0.01),
            txt='MoS2_out_gs.txt')

structure.calc = calc
structure.get_potential_energy()
calc.write('MoS2_gs.gpw', 'all')

calc.diagonalize_full_hamiltonian()
calc.write('MoS2_fulldiag.gpw', 'all')

for ecut in [80]:
    gw = G0W0(calc='MoS2_fulldiag.gpw',
              bands=(8, 18),
              ecut=ecut,
              truncation='2D',
              nblocksmax=True,
              q0_correction=True,
              filename=f'MoS2_g0w0_{ecut}',
              savepckl=True)

    gw.calculate()
